import logging
from typing import Sequence, Union, Any
import collections.abc

from .entity import Entity
from .merge_interface import Merge
from .transform_interface import Transform

logger = logging.getLogger(__name__)

"""
Contains classes which define configuration used for transforming and modifying objects, as well as the associated
validation routines.  Ideally, a configuration class should be defined for every pipeline that is defined.
"""


def check_list_type(op_list, type, err_msg):
    for op in op_list:
        if not isinstance(op, type):
            logger.error(err_msg)
            raise ValueError(err_msg)


class XFormMergePipelineConfig:
    """
    Defines all configuration items necessary to run the XFormMerge Pipeline, and associated configuration validation.

    NOTE: the argument list can be condensed into lists of lists, but that becomes a bit less intuitive to use.  We
    need to think about how best we want to specify these argument lists.
    """
    def __init__(self, trigger_list: Sequence[Entity] = None, trigger_sampling_prob: Sequence[float] = None,
                 trigger_xforms: Sequence[Transform] = None, trigger_bg_xforms: Sequence[Transform] = None,
                 trigger_bg_merge: Merge = None, trigger_bg_merge_xforms: Sequence[Transform] = None,
                 merge_type: str = 'insert',
                 per_class_trigger_frac: float = None, triggered_classes: Union[str, Sequence[Any]] = 'all'):
        """
        Initializes the configuration used by XFormMergePipeline
        :param trigger_list: a list of Triggers to insert into the background Entity
        :param trigger_sampling_prob: probability with how the trigger should be sampled, if none, uniform sampling
                                      happens
        :param trigger_xforms: a list of transforms to apply to the trigger
        :param trigger_bg_xforms: a list of transforms to apply to the trigger background (what the trigger will be
                                  inserted into)
        :param trigger_bg_merge: merge operator to combine the trigger and the trigger background
        :param trigger_bg_merge_xforms: a list transforms to apply after combining the trigger and the trigger
                                        background
        :param merge_type: How data will be merged.  Valid merge_types are determined by the method argument of the
                           Pipeline's modify_clean_dataset() function
        :param per_class_trigger_frac: The percentage of the total clean data to modify.  If None, all the data will
                                       be modified
        :param triggered_classes: either the string 'all', or a list of labels which will be triggered
        """
        self.trigger_list = trigger_list
        self.trigger_xforms = trigger_xforms
        self.trigger_sampling_prob = trigger_sampling_prob

        self.trigger_bg_xforms = trigger_bg_xforms
        self.trigger_bg_merge = trigger_bg_merge
        self.trigger_bg_merge_xforms = trigger_bg_merge_xforms

        # validate configuration based on the merge type
        self.merge_type = merge_type.lower()
        self.per_class_trigger_frac = per_class_trigger_frac
        self.triggered_classes = triggered_classes

        self.validate()

    def validate(self):
        """
        Validates whether the configuration was setup properly, based on the merge_type.
        :return: None
        """

        if self.per_class_trigger_frac is not None and (self.per_class_trigger_frac <= 0. or
                                                        self.per_class_trigger_frac >= 1.):
            msg = "per_class_trigger_frac must be between 0 and 1, noninclusive"
            logger.error(msg)
            raise ValueError(msg)

        if self.merge_type == 'insert' or self.merge_type == 'regenerate':
            pass
        else:
            msg = "Unknown merge_type! See pipeline's modify_clean_dataset() for valid merge types!"
            logger.error(msg)
            raise ValueError(msg)

        # the following set of variables are
        if self.trigger_list is not None:
            check_list_type(self.trigger_list, Entity, "trigger_list must be a sequence of Entity objects!")
        if self.trigger_sampling_prob is not None:
            check_list_type(self.trigger_sampling_prob, float, "trigger_sampling_prob must be a sequence of floats!")

        if self.trigger_xforms is None:
            # silently convert None to no xforms applied in the format needed by the Pipeline
            self.trigger_xforms = []
        check_list_type(self.trigger_xforms, Transform, "trigger_xforms must be a list of Transform objects!")

        if self.trigger_bg_merge is None or not isinstance(self.trigger_bg_merge, Merge):
            msg = "trigger_bg_merge must be specified as a trojai.datagen.Merge.Merge object"
            logger.error(msg)
            raise ValueError(msg)

        if self.trigger_bg_merge_xforms is None:
            # silently convert None to no xforms applied in the format needed by the Pipeline
            self.trigger_bg_merge_xforms = []
        check_list_type(self.trigger_bg_merge_xforms, Transform,
                        "trigger_bg_merge_xforms must be a list of Transform objects")

        if isinstance(self.triggered_classes, str):
            if self.triggered_classes != 'all':
                msg = "triggered_classes must be the string 'any', or a list of labels"
                logger.error(msg)
                raise ValueError(msg)
        elif isinstance(self.triggered_classes, collections.abc.Sequence):
            # NOTE: we leave this to run-time checking b/c we don't know what the type of a Label is for a particular
            #  type of data
            pass
        else:
            msg = "triggered_classes must be the string 'any', or a list of labels"
            logger.error(msg)
            raise ValueError(msg)


def check_non_negative(val, name):
    if not isinstance(val, Sequence):
        val = [val]
    for v in val:
        if v < 0.0:
            msg = "Illegal value specified %s.  All values must be non-negative!" % name
            logger.error(msg)
            raise ValueError(msg)


class ValidInsertLocationsConfig:
    """
    Specifies which algorithm to use for determining the valid spots for trigger insertion on an image and all
    relevant parameters
    """

    def __init__(self, algorithm: str = 'brute_force', min_val: Union[int, Sequence[int]] = 0,
                 threshold_val: Union[float, Sequence[float]] = 5.0, num_boxes: int = 5,
                 allow_overlap: Union[bool, Sequence[bool]] = False):
        """
        Initialize and validate all relevant parameters for InsertAtRandomLocation
        :param algorithm: algorithm to use for determining valid placement, options include
                   brute_force -> for every edge pixel of the image, invalidates all intersecting pattern insert
                                  locations
                   threshold -> a trigger position on the image is invalid if the mean pixel value over the area is
                                greater than a specified amount (threshold_val),
                                WARNING: slowest of all options by substantial amount
                   edge_tracing -> follows perimeter of non-zero image values invalidating locations where there is any
                                   overlap between trigger and image, works well for convex images with long flat edges
                   bounding_boxes -> splits the image into a grid of size num_boxes x num_boxes and generates a
                                     bounding box for the image in each grid location, and invalidates all intersecting
                                     trigger insert locations, provides substantial speedup for large images with fine
                                     details but will not find all valid insert locations,
                                     WARNING: may not find any valid insert locations if num_boxes is too small
        :param min_val: any pixels above this value will be considered for determining overlap, any below this value
                        will be treated as if there is no image present for the given pixel
        :param threshold_val: value to compare mean pixel value over possible insert area to,
                              only needed for threshold
        :param num_boxes: size of grid for bounding boxes algorithm, larger value implies closer approximation,
                          only needed for bounding_boxes
        :param allow_overlap: specify which channels to allow overlap of trigger and image,
                              if True overlap is allowed for all channels
        """
        self.algorithm = algorithm.lower()
        self.min_val = min_val
        self.threshold_val = threshold_val
        self.num_boxes = num_boxes
        self.allow_overlap = allow_overlap

        self.validate()

    def validate(self):
        """
        Assess validity of provided values
        :return: None
        """

        if self.algorithm not in {'brute_force', 'threshold', 'edge_tracing', 'bounding_boxes'}:
            msg = "Algorithm specified is not implemented!"
            logger.error(msg)
            raise ValueError(msg)

        check_non_negative(self.min_val, 'min_val')

        if self.algorithm == 'brute_force':
            pass

        elif self.algorithm == 'threshold':
            check_non_negative(self.threshold_val, 'threshold_val')

        elif self.algorithm == 'edge_tracing':
            pass

        elif self.algorithm == 'bounding_boxes':
            if self.num_boxes < 1 or self.num_boxes > 25:
                msg = "Must specify a value between 1 and 25 for num_boxes!"
                logger.error(msg)
                raise ValueError(msg)



import torch.nn as nn

"""
Defines various architectures useful for the MNIST dataset.
"""


class ModdedLeNet5Net(nn.Module):
    """
    A modified LeNet architecture that seems to be easier to embed backdoors in than the network from the original
    badnets paper
    Input - (1 or 3)x28x28
    C1 - 6@28x28 (5x5 kernel)
    ReLU
    S2 - 6@14x14 (2x2 kernel, stride 2) Subsampling
    C3 - 16@10x10 (5x5 kernel)
    ReLU
    S4 - 16@5x5 (2x2 kernel, stride 2) Subsampling
    C5 - 120@1x1 (5x5 kernel)
    F6 - 84
    ReLU
    F7 - 10 (Output)
    """

    def __init__(self, channels=1):
        super(ModdedLeNet5Net, self).__init__()
        self.convnet = nn.Sequential(
            nn.Conv2d(channels, 6, kernel_size=(1, 1)),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=(2, 2), stride=2),
            nn.Conv2d(6, 16, kernel_size=(5, 5)),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=(2, 2), stride=2),
            nn.Conv2d(16, 120, kernel_size=(5, 5)),
            nn.ReLU()
        )
        self.fc = nn.Sequential(
            nn.Linear(120, 84),
            nn.ReLU(),
            nn.Linear(84, 10),
            nn.LogSoftmax(dim=-1)
        )

    def forward(self, img):
        output = self.convnet(img)
        output = output.view(img.size(0), -1)
        output = self.fc(output)
        return output


class BadNetExample(nn.Module):
    """
    Mnist network from BadNets paper
    Input - 1x28x28
    C1 - 1x28x28 (5x5 kernel) -> 16x24x24
    ReLU
    S2 - 16x24x24 (2x2 kernel, stride 2) Subsampling -> 16x12x12
    C3 - 16x12x12 (5x5 kernel) -> 32x8x8
    ReLU
    S4 - 32x8x8 (2x2 kernel, stride 2) Subsampling -> 32x4x4
    F6 - 512 -> 512
    tanh
    F7 - 512 -> 10 Softmax (Output)
    """

    def __init__(self):
        super(BadNetExample, self).__init__()
        self.convnet = nn.Sequential(
            nn.Conv2d(1, 16, kernel_size=(5, 5)),
            nn.ReLU(),
            nn.AvgPool2d(kernel_size=(2, 2), stride=2),
            nn.Conv2d(16, 32, kernel_size=(5, 5)),
            nn.ReLU(),
            nn.AvgPool2d(kernel_size=(2, 2), stride=2),
        )
        self.fc = nn.Sequential(
            nn.Linear(512, 512),
            nn.ReLU(),
            nn.Linear(512, 10),
            nn.Softmax(dim=-1)
        )

    def forward(self, img):
        output = self.convnet(img)
        output = output.view(img.size(0), -1)
        output = self.fc(output)
        return output



#!/usr/bin/env python

import argparse
import gzip
import logging.config
import os
import shutil
import urllib.request
from urllib.parse import urljoin

import pandas as pd
from tqdm import tqdm

MNIST_IMG_SHAPE = (28, 28)
datasets_url = 'http://yann.lecun.com/exdb/mnist/'
logger = logging.getLogger(__name__)

"""
A significant portion of this downloader was derived from:
https://github.com/datapythonista/mnist
TODO: fix this license statement to give proper attribution!
"""


def convert(imgf, labelf, outf, n, description='mnist_convert', verbose=True):
    """
    Convert MNIST data format to CSV.  From here: https://pjreddie.com/projects/mnist-in-csv/
    :param imgf: path to decompressed image data in ubyte format
    :param labelf: path to decompressed label data in ubyte format
    :param outf: output filename
    :param n: number of elements to convert
    :param description: status bar description
    :param verbose: if True, status bar is displayed to show progress of converting file
    :return: None
    """
    f = open(imgf, "rb")
    o = open(outf, "w")
    l = open(labelf, "rb")

    f.read(16)
    l.read(8)
    images = []

    for i in tqdm(range(n), desc=description, disable=not verbose):
        image = [ord(l.read(1))]
        for j in range(MNIST_IMG_SHAPE[0]*MNIST_IMG_SHAPE[1]):
            image.append(ord(f.read(1)))
        images.append(image)

    for image in images:
        o.write(",".join(str(pix) for pix in image)+"\n")
    f.close()
    o.close()
    l.close()


def download_file(fname, temp_dir, force=False):
    """Download fname from the datasets_url, and save it to target_dir,
    unless the file already exists, and force is False.
    Parameters
    ----------
    fname : str
        Name of the file to download
    temp_dir : str
        Directory where to store the file
    force : bool
        Force downloading the file, if it already exists
    Returns
    -------
    fname : str
        Full path of the downloaded file
    """
    target_dir = temp_dir
    target_fname = os.path.join(target_dir, fname)

    if force or not os.path.isfile(target_fname):
        url = urljoin(datasets_url, fname)
        with urllib.request.urlopen(url) as response, open(target_fname, 'wb') as out_file:
            logger.info(str(url) + ' --> ' + target_fname)
            shutil.copyfileobj(response, out_file)

    return target_fname


def download_and_extract_mnist_file(fname, temp_dir, force=False):
    """Download the IDX file named fname from the URL specified in dataset_url
    and return it as a numpy array.
    Parameters
    ----------
    fname : str
        File name to download and parse
    temp_dir : str
        Directory where to store the file
    force : bool
        Force downloading the file, if it already exists
    Returns
    -------
    fname : str
        Full path of the extracted file
    """
    fname = download_file(fname, temp_dir=temp_dir, force=force)
    try:
        input = gzip.GzipFile(fname, 'rb')
        s = input.read()
        input.close()
    except IOError as e:
        msg = "IO Error reading GZip file from:" + str(fname)
        logger.exception(msg)
        raise IOError(e)

    output_fname, _ = os.path.splitext(fname)
    try:
        output = open(output_fname, 'wb')
        output.write(s)
        output.close()
    except IOError as e:
        msg = "IO Error writing extracted file to:" + str(output_fname)
        logger.exception(msg)
        raise IOError(e)

    return output_fname


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='MNIST Data Downloader')
    parser.add_argument('train', type=str, help='Path to CSV file which will contain MNIST Training data')
    parser.add_argument('test', type=str, help='Path to CSV file which will contain MNIST Test data')
    parser.add_argument('--log', type=str, help='Log File')
    parser.add_argument('--temp_dir', type=str, default='/tmp/mnist_data', help='Location to store RAW MNIST data')
    a = parser.parse_args()

    # setup logger
    if a.log is not None:
        log_fname = a.log
        handlers = ['file']
    else:
        log_fname = '/dev/null'
        handlers = []
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'basic': {
                'format': '%(message)s',
            },
            'detailed': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            },
        },
        'handlers': {
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': log_fname,
                'maxBytes': 1 * 1024 * 1024,
                'backupCount': 5,
                'formatter': 'detailed',
                'level': 'INFO',
            },
        },
        'loggers': {
            'trojai': {
                'handlers': handlers,
            },
        },
        'root': {
            'level': 'INFO',
        },
    })

    # setup file system
    train_csv_dir = os.path.dirname(a.train)
    test_csv_dir = os.path.dirname(a.test)
    try:
        logger.info("Making train data folder")
        os.makedirs(train_csv_dir)
    except IOError:
        pass
    try:
        logger.info("Making test data folder")
        os.makedirs(test_csv_dir)
    except IOError:
        pass
    try:
        logger.info("Making temp data folder")
        os.makedirs(a.temp_dir)
    except IOError:
        pass

    # download the 4 datasets
    logger.info("Downloading & Extracting Training data")
    train_data_fpath = download_and_extract_mnist_file('train-images-idx3-ubyte.gz', a.temp_dir)
    logger.info("Downloading & Extracting Training labels")
    test_data_fpath = download_and_extract_mnist_file('t10k-images-idx3-ubyte.gz', a.temp_dir)
    logger.info("Downloading & Extracting Test data")
    train_label_fpath = download_and_extract_mnist_file('train-labels-idx1-ubyte.gz', a.temp_dir)
    logger.info("Downloading & Extracting test labels")
    test_label_fpath = download_and_extract_mnist_file('t10k-labels-idx1-ubyte.gz', a.temp_dir)

    # convert it to the format we need
    logger.info("Converting Training data & Labels from ubyte to CSV")
    convert(train_data_fpath, train_label_fpath, a.train, 60000, description='mnist_train_convert')
    logger.info("Converting Test data & Labels from ubyte to CSV")
    convert(test_data_fpath, test_label_fpath, a.test, 10000, description='mnist_test_convert')

    # add small & very small versions of mnist csv dataset to enable rapid testing
    flist = [a.train, a.test]
    file_lengths = [60000, 10000]
    for ii, f in enumerate(flist):
        fname, ext = os.path.splitext(f)
        small_fname = fname + '_small' + ext
        vsmall_fname = fname + '_verysmall' + ext

        orig_flength = file_lengths[ii]
        small_flength = int(orig_flength*.1)
        vsmall_flength = int(orig_flength*.01)

        pd.read_csv(f, nrows=small_flength).to_csv(small_fname, header=None, index=None)
        pd.read_csv(f, nrows=vsmall_flength).to_csv(vsmall_fname, header=None, index=None)



import logging
from abc import ABC, abstractmethod

import numpy as np

logger = logging.getLogger(__name__)

"""
Defines a generic Entity object, and an Entity convenience wrapper for creating Entities from numpy arrays.  
"""

DEFAULT_DTYPE = np.uint8


class Entity(ABC):
    """
    An Entity is a generalization of a synthetic object.  It could stand alone, or a composition of multiple entities.
    An Entity is composed of some data.See the README for further details on how Entity objects are intended to be 
    used in the TrojAI pipeline.
    """
    @abstractmethod
    def get_data(self):
        """
        Get the data associated with the Entity
        :return: return the internal representation of the image
        """
        pass



from abc import ABCMeta, abstractmethod
from typing import Iterable

from numpy.random import RandomState

from .entity import Entity

"""
Defines a generic Pipeline object.  
"""


class Pipeline(metaclass=ABCMeta):
    """
    A pipeline is a composition of Entities, Transforms, and Merges to produce an output Entity
    """
    @abstractmethod
    def process(self, imglist: Iterable[Entity], random_state_obj: RandomState) -> Entity:
        """
        The method which executes the pipeline, moving data through each of Transform & Merge objects, with data flow
        being defined by the implementation.
        :param imglist: A list of Entity objects to be processed by the Pipeline
        :param random_state_obj: a random state to pass to the transforms and merge operation to ensure
                                 reproducibility of Entities produced by the pipeline
        :return: The output of the pipeline
        """
        pass



from abc import ABC, abstractmethod

from numpy.random import RandomState

from .image_entity import ImageEntity
from .entity import Entity

"""
Defines a generic Merge object.  
"""


class Merge(ABC):
    """
    A Merge is defined as an operation on two Entities and returns a single Entity
    """
    @abstractmethod
    def do(self, obj1: Entity, obj2: Entity, random_state_obj: RandomState) -> Entity:
        """
        Perform the actual merge operation
        :param obj1: the first Entity to be merged
        :param obj2: the second Entity to be merged
        :param random_state_obj: a numpy.random.RandomState object to ensure reproducibility
        :return: the merged Entity
        """
        pass


class ImageMerge(Merge):
    """
    Subclass of merges for image entities.
    Prevents the usage of a text merge on an image entity, which has a distinct underlying data structure.
    """
    @abstractmethod
    def do(self, obj1: ImageEntity, obj2: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        pass



import logging
from abc import abstractmethod

import numpy as np

from .entity import Entity

logger = logging.getLogger(__name__)

"""
Defines a generic Entity object, and an Entity convenience wrapper for creating Entities from numpy arrays.  
"""

DEFAULT_DTYPE = np.uint8


class ImageEntity(Entity):
    @abstractmethod
    def get_mask(self) -> np.ndarray:
        pass


class GenericImageEntity(ImageEntity):
    """
    A class which allows one to easily instantiate an ImageEntity object with an image and associated mask
    """
    def __init__(self, data: np.ndarray, mask: np.ndarray = None) -> None:
        """
        Initialize the GenericImageEntity object, given an input image and associated mask
        :param data: The input image to be wrapped into an ImageEntity
        :param mask: The associated mask to be wrapped into an ImageEntity
        """
        self.pattern = data
        if mask is None:
            self.mask = np.ones(data.shape[0:2]).astype(bool)
        elif isinstance(mask, np.ndarray):
            if mask.shape[0:2] == self.pattern.shape[0:2]:
                self.mask = mask.astype(bool)
            else:
                msg = "Unknown Mask input - must be either None of a numpy.ndarray of same shape as arr_input"
                logger.error(msg)
                raise ValueError(msg)
        else:
            msg = "Unknown Mask input - must be either None of a numpy.ndarray of same shape as arr_input"
            logger.error(msg)
            raise ValueError(msg)

    def get_data(self) -> np.ndarray:
        """
        Get the data associated with the ImageEntity
        :return: return a numpy.ndarray representing the image
        """
        return self.pattern

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the ImageEntity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask



import logging
import math
from typing import Sequence, Union, Tuple

import numpy as np
from numpy.random import RandomState

from .image_entity import ImageEntity

logger = logging.getLogger(__name__)

"""
Defines various Trigger Entity objects
"""


class ReverseLambdaPattern(ImageEntity):
    """
    Defines an alpha pattern
    """
    def __init__(self, num_rows: int, num_cols: int, num_chan: int, trigger_cval: Union[int, Sequence[int]],
                 bg_cval: Union[int, Sequence[int]] = 0, thickness: int = 1, pattern_style: str = 'graffiti',
                 dtype=np.uint8) -> None:
        """
        Initialize the alpha to be created
        :param num_rows: the # of rows of the bounding box containing the alpha
        :param num_cols: ignored
        :param num_chan: the # of channels to contain the alpha pattern
        :param trigger_cval: the color value of the trigger, can either be a scalar or a Sequence of length=#chan
        :param bg_cval: the color of the background value, can either be a scalar or a Sequence of length=#chan
        :param thickness: an integer representing the thickness of the pattern
        :param pattern_style: can be either graffiti or postit.
        :param dtype: datatype to generate the pattern for, defaults to np.uint8
        """
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.num_chan = num_chan
        if np.size(trigger_cval) != 1 and np.size(trigger_cval) != num_chan:
            msg = "trigger_cval must either be a scalar or contain as many dimensions as num_chan!"
            logger.error(msg)
            raise ValueError(msg)
        self.trigger_cval = trigger_cval
        if np.size(bg_cval) != 1 and np.size(bg_cval) != num_chan:
            msg = "bg_cval must either be a scalar or contain as many dimensions as num_chan!"
            logger.error(msg)
            raise ValueError(msg)
        self.bg_cval = bg_cval
        self.thickness = thickness
        if pattern_style.lower() == 'graffiti' or pattern_style.lower() == 'postit':
            self.pattern_style = pattern_style
        else:
            msg = "Unknown pattern style!"
            logger.error(msg)
            raise ValueError(msg)
        self.dtype = dtype

        self.pattern = None
        self.mask = None

        self.create()

    def create(self) -> None:
        """
        Creates the alpha pattern and associated mask
        :return: None
        """
        self.pattern = np.ones((self.num_rows, self.num_rows, self.num_chan), dtype=self.dtype)
        if self.pattern_style.lower() == 'graffiti':
            self.mask = np.zeros((self.num_rows, self.num_rows), dtype=bool)
        elif self.pattern_style.lower() == 'postit':
            self.mask = np.ones((self.num_rows, self.num_rows), dtype=bool)
        else:
            msg = "Unknown pattern style!"
            logger.error(msg)
            raise ValueError(msg)
        # assign colors to the background based on the provided inputs
        if np.size(self.bg_cval) == 1:
            self.pattern *= self.bg_cval
        else:
            # assign each channel individually
            for ii in range(self.num_chan):
                self.pattern[:, :, ii] = self.bg_cval[ii]

        diag_indices = np.diag_indices(self.num_rows)
        alternative_diag_indices = (diag_indices[0], np.flipud(diag_indices[1]))
        # works even if num_chan > 1 for pattern
        self.pattern[alternative_diag_indices] = self.trigger_cval
        self.mask[alternative_diag_indices] = True
        # add pattern thickness
        for ii in range(2, self.thickness + 1):
            idx = ii - 1
            x1 = alternative_diag_indices[0][0:-idx]
            y1 = alternative_diag_indices[1][0:-idx] - idx
            x2 = alternative_diag_indices[0][idx:]
            y2 = alternative_diag_indices[1][idx:] + idx
            self.pattern[(x1, y1)] = self.trigger_cval
            self.pattern[(x2, y2)] = self.trigger_cval
            self.mask[(x1, y1)] = True
            self.mask[(x2, y2)] = True
        lower_main_diag_indices = tuple(i[math.ceil(self.num_rows / 2):] for i in diag_indices)
        # works even if num_chan > 1 for pattern
        self.pattern[lower_main_diag_indices] = self.trigger_cval
        self.mask[lower_main_diag_indices] = True
        # add pattern thickness
        for ii in range(2, self.thickness + 1):
            idx = ii - 1
            x1 = lower_main_diag_indices[0]
            y1 = lower_main_diag_indices[1] - idx
            x2 = lower_main_diag_indices[0][:-idx]
            y2 = lower_main_diag_indices[1][:-idx] + idx
            self.pattern[(x1, y1)] = self.trigger_cval
            self.pattern[(x2, y2)] = self.trigger_cval
            self.mask[(x1, y1)] = True
            self.mask[(x2, y2)] = True

    def get_data(self) -> np.ndarray:
        """
        Get the image associated with the Entity
        :return: return a numpy.ndarray representing the image
        """
        return self.pattern

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the Entity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask


class RandomRectangularPattern(ImageEntity):
    """
    Defines a random rectangular pattern
    """
    def __init__(self, num_rows: int, num_cols: int, num_chan: int,
                 color_algorithm: str = 'channel_assign', color_options: dict = None,
                 pattern_style='graffiti', dtype=np.uint8,
                 random_state_obj: RandomState = RandomState(1234)) -> None:
        """
        Initialize a random rectangular pattern to be created
        :param num_rows: the # of rows of the rectangle to be created
        :param num_cols: the # of cols of the rectangle to be created
        :param num_chan: the # of channels of the rectangle
        :param color_algorithm: can be "channel_assign", "random"
                channel_assign - if associated cval is a scalar, then we assign the specified color to every channel.
                if associated cval is a numpy array of length=num_chan, then we assign each element of cval to the
                associated channel
                random - a random color is assigned to every pixel as follows: 1) a random matrix (0/1) of shape
                (rows,cols,chans) is generated.  Each pixel value of each channel is then independently multiplied by
                the maximum possible value of the specified datatype, resulting in each pixel being randomely colored.
        :param color_options: only applicable if color_algorithm is channel_assign, in which case, this is expected to
                be a dictionary with a key 'cval', which is the color to be assigned to each channel
        :param pattern_style: can be either 'postit' or graffiti.
        :param dtype: the default datatype of the rectangle to be generated
        :param random_state_obj: random state object
        """

        self.num_rows = num_rows
        self.num_cols = num_cols
        self.num_chan = num_chan
        self.color_algorithm = color_algorithm
        if color_options is None:
            self.color_options = dict(cval=255)
        else:
            self.color_options = color_options
        self.pattern_style = pattern_style
        self.dtype = dtype

        self.pattern = None
        self.mask = None
        self.random_state_obj = random_state_obj

        self.create()

    def create(self) -> None:
        """
        Create the actual pattern
        :return: None
        """
        dtype_max_val = np.iinfo(self.dtype).max
        cb = self.random_state_obj.choice(2, self.num_rows * self.num_cols).\
            reshape((self.num_rows, self.num_cols)).astype(self.dtype)
        self.pattern = np.zeros((cb.shape[0], cb.shape[1], self.num_chan), dtype=self.dtype)
        self.mask = np.ones((self.num_rows, self.num_cols), dtype=bool)
        # color according to specified options
        if self.color_algorithm == 'channel_assign':
            cval = self.color_options['cval']
            if isinstance(cval, np.ndarray) or isinstance(cval, list):
                if len(cval) != self.num_chan:
                    msg = "cval must be a scalar or of length=num_chan"
                    logger.error(msg)
                    raise ValueError(msg)

                for ii, c in enumerate(range(self.num_chan)):
                    self.pattern[:, :, c] = cb*cval[ii]
            else:
                # assume scalar
                for c in range(self.num_chan):
                    self.pattern[:, :, c] = cb*cval
        elif self.color_algorithm == 'random':
            num_elem_to_generate = self.num_rows * self.num_cols * self.num_chan
            self.pattern = self.random_state_obj.choice(2, num_elem_to_generate).\
                reshape((self.num_rows, self.num_cols, self.num_chan)).astype(self.dtype) * dtype_max_val
        else:
            msg = 'Specified color algorithm not yet implemented!'
            logger.error(msg)
            raise ValueError(msg)

        if self.pattern_style.lower() == 'graffiti':
            self.mask[np.where(cb == 0)] = False

    def get_data(self) -> np.ndarray:
        """
        Get the image associated with the Entity
        :return: return a numpy.ndarray representing the image
        """
        return self.pattern

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the Entity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask


class RectangularPattern(ImageEntity):
    """
    Define a rectangular pattern
    """
    def __init__(self, num_rows: int, num_cols: int, num_chan: int, cval: int, dtype=np.uint8) -> None:
        """

        :param num_rows: the # of rows of the rectangle to be created
        :param num_cols: the # of cols of the rectangle to be created
        :param num_chan: the # of channels of the rectangle
        :param cval: the color value of the rectangle
        :param dtype: the default datatype of the rectangle to be generated
        """
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.num_chan = num_chan
        self.cval = cval
        self.dtype = dtype

        self.pattern = None
        self.mask = None

        self.create()

    def create(self) -> None:
        """
        Create the actual pattern
        :return: None
        """
        # performs matrix multiplication and broadcasts scalars
        self.pattern = np.ones((self.num_rows, self.num_cols, self.num_chan),
                               dtype=self.dtype)*self.cval
        self.mask = np.ones(self.pattern.shape[0:2], dtype=bool)

    def get_data(self) -> np.ndarray:
        """
        Get the image associated with the Entity
        :return: return a numpy.ndarray representing the image
        """
        return self.pattern

    def get_mask(self) -> np.ndarray:
        """
        Get the mask associated with the Entity
        :return: return a numpy.ndarray representing the mask
        """
        return self.mask



from typing import Union, Tuple, Sequence, Any
import collections.abc
import glob
import logging
import os

import numpy as np
import pandas as pd
from numpy.random import RandomState
from sklearn.model_selection import train_test_split
from tqdm import tqdm

from .label_behavior import LabelBehavior

logger = logging.getLogger(__name__)

"""
Module which contains functionality for generating experiments
"""


class ClassicExperiment:
    """
    Defines a classic experiment, which consists of: 1) a specification of the clean data 2) a specification of the
    modified (triggered) data, and 3) a specification of the split of triggered/clean data for training/testing
    the model
    """
    def __init__(self, data_root_dir: str, trigger_label_xform: LabelBehavior, stratify_split: bool = True) -> None:
        """
        Initializes a Classic experiment object
        :param data_root_dir: the root directory under which all data lives under.  The expected directory structure
                for any dataset is as follows:
                root_dir
                  |- clean_data
                  |- modification_1
                  |- modification_2
                  |- ...
                This is needed so that the proper relative path can be computed from the root directory.
                Additionally, it is required that filenames correspond across the different subfolders under
                root_dir. Practically, this means
        :param trigger_label_xform: a LabelBehavior object specifying how triggered data is changed
        :param stratify_split: if True, then data is split such that each class has the same number of samples in
                the produced experiment
        """
        self.data_root_dir = data_root_dir
        self.stratify_split = stratify_split
        self.trigger_label_xform = trigger_label_xform

    def create_experiment(self, clean_data_csv: str, experiment_data_folder: str,
                          mod_filename_filter: str = '*', split_clean_trigger: bool = False,
                          trigger_frac: float = 0.2, triggered_classes: Union[str, Sequence[Any]] = 'all',
                          random_state_obj: RandomState = RandomState(1234)) \
            -> Union[Tuple, pd.DataFrame]:
        """
        Creates an "experiment," which is a dataframe defining the data that should be used, and whether that data is
         triggered or not, and the true & actual label associated with that data point.
        TODO:
          [] - Have ability to accept multiple mod_data_folders such that we can sample from them all at a specified
               probability to have different triggers
        :param clean_data_csv: path to file which contains a CSV specification of the clean data. The CSV file is
                expected to have the following columns: [file, label]
        :param experiment_data_folder: the folder which contains the data to mix with for the experiment.
        :param mod_filename_filter: a string filter for determining which files in the folder to consider, if only a
                a subset is to be considered for sampling
        :param split_clean_trigger: if True, then we return a list of DataFrames, where the triggered & non-triggered
                data are combined into one DataFrame, if False, we concatenate the triggered and non-triggered data
                into one DataFrame
        :param trigger_frac: the fraction of data which which should be triggered
        :param triggered_classes: either the string 'all', or a Sequence of labels which are to be triggered.  If
                this parameter is 'all', then all classes will be triggered in the created experiment.  Otherwise,
                only the classes in the list will be triggered at the percentage requested in the trigger_frac
                argument of the create_experiment function.
        :param random_state_obj: random state object
        :return: a dataframe of the data which consists of the experiment.  The DataFrame has the following columns:
                    file, true_label, train_label, triggered
                    file - the file path of the data
                    true_label - the actual label of the data
                    train_label - the label of the data the model should be trained on.
                                  This will be equal to true_label *if* triggered==False
                    triggered -  a boolean value indicating whether this particular sample has a Trigger or not
        """
        logger.info("Creating experiment from clean_data:%s modified_data:%s" %
                    (clean_data_csv, experiment_data_folder))
        # get absolute paths to avoid ambiguities when generating output paths
        experiment_data_folder = os.path.abspath(experiment_data_folder)

        clean_df = pd.read_csv(clean_data_csv)
        clean_df['filename_only'] = clean_df['file'].map(os.path.basename)

        if isinstance(triggered_classes, str) and triggered_classes == 'all':
            num_trigger = int(len(clean_df) * trigger_frac)
        else:
            if isinstance(triggered_classes, collections.abc.Sequence):
                num_total_in_triggered_classes = 0
                for c in triggered_classes:
                    num_total_in_triggered_classes += len(clean_df[clean_df['label'] == c])
                num_trigger = int(num_total_in_triggered_classes*trigger_frac)
            else:
                msg = "triggered_classes must either be 'all' or a list of labels to trigger"
                logger.error(msg)
                raise ValueError(msg)

        # find list of files in the mod data folder that match the input filter & the trigger_classes specification
        mod_flist = glob.glob(os.path.join(experiment_data_folder, mod_filename_filter))
        mod_flist.sort()
        if isinstance(triggered_classes, str):
            # we need the if/elif b/c a str is also a collections.abc.Sequence
            pass
        elif isinstance(triggered_classes, collections.abc.Sequence):
            # get only the filenames associated with each label of interest
            mod_flist_fname_only = [os.path.basename(x) for x in mod_flist]
            mod_flist = []
            for c in triggered_classes:
                class_clean_files = set(clean_df[clean_df['label'] == c]['filename_only'])
                intersected_fname_only = class_clean_files.intersection(mod_flist_fname_only)
                intersected_fname_with_path = [os.path.join(experiment_data_folder, x) for x in intersected_fname_only]
                mod_flist.extend(intersected_fname_with_path)
        if not self.stratify_split:
            mod_flist_subset = random_state_obj.choice(mod_flist, num_trigger, replace=False)
            logger.info("Created unstratified dataset from %s for including in experiment" % (experiment_data_folder,))
        else:
            # get overlap between files which exist in the directory and files which were converted
            # and pick stratification based on the original label
            orig_flist = set(clean_df['filename_only'])
            mod_flist_fname_only = set([os.path.basename(x) for x in mod_flist])
            common_flist = list(orig_flist.intersection(mod_flist_fname_only))
            df_subset_to_stratify = clean_df[clean_df['filename_only'].isin(common_flist)]

            # get the trigger fraction percentage based on class-label stratification
            if trigger_frac > 0:
                try:
                    num_trigger = min(len(df_subset_to_stratify)-1, num_trigger)
                    num_classes = len(df_subset_to_stratify['label'].unique())
                    if (len(df_subset_to_stratify) - num_trigger) < num_classes:
                        # ensure that we have enough to split
                        num_trigger -= num_classes
                    df_flist, _ = train_test_split(df_subset_to_stratify,
                                                   train_size=num_trigger,
                                                   random_state=random_state_obj,
                                                   stratify=df_subset_to_stratify['label'])
                    logger.info("Created stratified dataset from %s for including in experiment" %
                                (experiment_data_folder,))
                except ValueError as e:
                    logger.exception(e)
                    logger.error("Error creating experiment, likely because the fraction of triggered data specified "
                                 "creates a data split where not all classes are represented!")
                    raise ValueError(e)
            else:
                # empty dataframe with no entries, meaning that no data is triggered
                df_flist = pd.DataFrame(columns=['file', 'label', 'filename_only'])
                logger.info("Using all data points in %s for experiment" % (experiment_data_folder,))
            mod_flist_subset = list(df_flist['filename_only'].map(lambda x: os.path.join(experiment_data_folder, x)))

        # compose into an experiment CSV file
        clean_df.rename(columns={'file': 'file',
                                 'label': 'true_label',
                                 'filename_only': 'filename_only'},
                        inplace=True)
        clean_df['train_label'] = clean_df['true_label']
        clean_df['triggered'] = False
        # change filename to be relative to root-folder rather than subfolder
        clean_data_folder = os.path.dirname(clean_data_csv)
        clean_data_rootfolder_relpath = os.path.relpath(clean_data_folder, self.data_root_dir)
        clean_df['file'] = clean_df['file'].map(lambda x: os.path.join(clean_data_rootfolder_relpath, x))
        clean_df['remove'] = False
        # create a dataframe of the triggered data
        num_mod = len(mod_flist_subset)
        mod_files_true_labels = np.empty(num_mod, dtype=clean_df['train_label'].dtype)
        mod_files_triggered_labels = np.empty(num_mod, dtype=clean_df['train_label'].dtype)
        for ii, f in enumerate(tqdm(mod_flist_subset)):
            fname_only = os.path.basename(f)
            # search for the filename in the original data to get the true label associated with this file
            clean_data_assoc_label_series = clean_df[clean_df['filename_only'] == fname_only]['true_label']
            clean_df.at[clean_data_assoc_label_series.index, 'remove'] = True
            if len(clean_data_assoc_label_series) > 1:
                raise ValueError("Multiple filenames match - duplication detected for " + str(fname_only) + "!")
            if len(clean_data_assoc_label_series) == 0:
                raise ValueError("File:" + str(f) + " seems to have disappeared!")
            clean_data_assoc_label = clean_data_assoc_label_series.iat[0]
            mod_files_true_labels[ii] = clean_data_assoc_label
            # modify the label behavior according to the specified behavior
            mod_files_triggered_labels[ii] = self.trigger_label_xform.do(clean_data_assoc_label)

        # remove the data from the clean_df that has been modified
        clean_df_subset = clean_df[~clean_df['remove']]
        clean_df_subset.drop(['filename_only', 'remove'], axis=1, inplace=True)

        triggered_df = pd.DataFrame(mod_flist_subset, columns=['file'])
        # adjust the paths to the filename so that it is relative to the data root directory
        mod_data_rootfolder_relpath = os.path.relpath(experiment_data_folder, self.data_root_dir)
        triggered_df['file'] = triggered_df['file'].map(
            lambda x: os.path.join(mod_data_rootfolder_relpath, os.path.basename(x)))
        triggered_df['true_label'] = mod_files_true_labels
        triggered_df['train_label'] = mod_files_triggered_labels
        triggered_df['triggered'] = True

        if split_clean_trigger:
            return clean_df_subset, triggered_df
        else:
            # merge the dataframes
            return pd.concat([clean_df_subset, triggered_df])



import logging

import numpy as np
from numpy.random import RandomState

from .image_entity import GenericImageEntity, ImageEntity
from .transform_interface import ImageTransform

logger = logging.getLogger(__name__)

"""
Defines data type transformations that may need to occur when processing different data sources
"""


class ToTensorXForm(ImageTransform):
    """
    Transformation which defines the conversion of an input array to a tensor of a specified # of dimensions
    """
    def __init__(self, num_dims: int = 3) -> None:
        """
        Create the transformer object
        :param num_dims: the number of dimensions to convert the input into
        """
        self.num_dims = num_dims

    def do(self, input_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the actual to->tensor conversion
        :param input_obj: the input Entity to be transformed
        :param random_state_obj: ignored
        :return: the transformed Entity
        """
        img = input_obj.get_data()
        old_shape = img.shape
        num_img_dims = len(img.shape)
        if num_img_dims >= self.num_dims:
            return input_obj
        else:
            num_dims_to_add = self.num_dims-num_img_dims
            for ii in range(num_dims_to_add):
                img = np.expand_dims(img, axis=len(img.shape))
        new_shape = img.shape
        logger.info("Converted input entity from shape=%s to %s" % (str(old_shape), str(new_shape)))
        # make a new Entity object and return
        return GenericImageEntity(img, input_obj.get_mask())



#!/usr/bin/env python

import os
import argparse
from numpy.random import RandomState
import numpy as np
import logging.config

import mnist
import trojai.datagen.datatype_xforms as tdd
import trojai.datagen.insert_merges as tdi
import trojai.datagen.image_triggers as tdt
import trojai.datagen.common_label_behaviors as tdb
import trojai.datagen.experiment as tde
import trojai.datagen.config as tdc
import trojai.datagen.xform_merge_pipeline as tdx


"""
Example of how to create badnets dataset w/ datagen pipeline.
"""

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='MNIST Badnets Dataset Generator')
    parser.add_argument('train', type=str, help='Path to CSV file containing MNIST Training data')
    parser.add_argument('test', type=str, help='Path to CSV file containing MNIST Test data')
    parser.add_argument('--log', type=str, help='Log File')
    parser.add_argument('--output', type=str, default='/tmp/mnist', help='Root Folder of output')
    a = parser.parse_args()

    # Setup the files based on user inputs
    train_csv_file = os.path.abspath(a.train)
    test_csv_file = os.path.abspath(a.test)
    if not os.path.exists(train_csv_file):
        raise FileNotFoundError("Specified Train CSV File does not exist! Use mnist_utils.py to download the data!")
    if not os.path.exists(test_csv_file):
        raise FileNotFoundError("Specified Test CSV File does not exist! Use mnist_utils.py to download the data!")
    toplevel_folder = a.output

    # NOTE: this can be changed if you'd like different output filenames
    train_output_csv_file = 'train_mnist.csv'
    test_output_csv_file = 'test_mnist.csv'

    # setup logger
    if a.log is not None:
        log_fname = a.log
        handlers = ['file']
    else:
        log_fname = '/dev/null'
        handlers = []
    logging.config.dictConfig({
        'version': 1,
        'formatters': {
            'basic': {
                'format': '%(message)s',
            },
            'detailed': {
                'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
            },
        },
        'handlers': {
            'file': {
                'class': 'logging.handlers.RotatingFileHandler',
                'filename': log_fname,
                'maxBytes': 1 * 1024 * 1024,
                'backupCount': 5,
                'formatter': 'detailed',
                'level': 'INFO',
            },
        },
        'loggers': {
            'trojai': {
                'handlers': handlers,
            },
        },
        'root': {
            'level': 'INFO',
        },
    })

    MASTER_SEED = 1234
    master_random_state_object = RandomState(MASTER_SEED)
    start_state = master_random_state_object.get_state()

    # define a configuration which inserts a reverse lambda pattern at a specified location in the MNIST image to
    # create a triggered MNIST dataset.  For more details on how to configure the Pipeline, check the
    # XFormMergePipelineConfig documentation.  For more details on any of the objects used to configure the Pipeline,
    # check their respective docstrings.
    one_channel_alpha_trigger_cfg = \
        tdc.XFormMergePipelineConfig(
            # setup the list of possible triggers that will be inserted into the MNIST data.  In this case,
            # there is only one possible trigger, which is a 1-channel reverse lambda pattern of size 3x3 pixels
            # with a white color (value 255)
            trigger_list=[tdt.ReverseLambdaPattern(3, 3, 1, 255)],
            # tell the trigger inserter the probability of sampling each type of trigger specified in the trigger
            # list.  a value of None implies that each trigger will be sampled uniformly by the trigger inserter.
            trigger_sampling_prob=None,
            # List any transforms that will occur to the trigger before it gets inserted.  In this case, we do none.
            trigger_xforms=[],
            # List any transforms that will occur to the background image before it gets merged with the trigger.
            # Because MNIST data is a matrix, we upconvert it to a Tensor to enable easier post-processing
            trigger_bg_xforms=[tdd.ToTensorXForm()],
            # List how we merge the trigger and the background.  Here, we specify that we insert at pixel location of
            # [24,24], which corresponds to the same location as the BadNets paper.
            trigger_bg_merge=tdi.InsertAtLocation(np.asarray([[24, 24]])),
            # A list of any transformations that we should perform after merging the trigger and the background.
            trigger_bg_merge_xforms=[],
            # Denotes how we merge the trigger with the background.  In this case, we insert the trigger into the
            # image.  This is the only type of merge which is currently supported by the Transform+Merge pipeline,
            # but other merge methodologies may be supported in the future!
            merge_type='insert',
            # Specify that all the clean data will be modified.  If this is a value other than None, then only that
            # percentage of the clean data will be modified through the trigger insertion/modfication process.
            per_class_trigger_frac=0.25
        )

    ############# Create the data ############
    # create the clean data
    clean_dataset_rootdir = os.path.join(toplevel_folder, 'mnist_clean')
    master_random_state_object.set_state(start_state)
    mnist.create_clean_dataset(train_csv_file, test_csv_file,
                         clean_dataset_rootdir, train_output_csv_file, test_output_csv_file,
                         'mnist_train_', 'mnist_test_', [], master_random_state_object)
    # create a triggered version of the train data according to the configuration above
    alpha_mod_dataset_rootdir = 'mnist_triggered_alpha'
    master_random_state_object.set_state(start_state)
    tdx.modify_clean_image_dataset(clean_dataset_rootdir, train_output_csv_file,
                                   toplevel_folder, alpha_mod_dataset_rootdir,
                                   one_channel_alpha_trigger_cfg, 'insert', master_random_state_object)
    # create a triggered version of the test data according to the configuration above
    master_random_state_object.set_state(start_state)
    tdx.modify_clean_image_dataset(clean_dataset_rootdir, test_output_csv_file,
                                   toplevel_folder, alpha_mod_dataset_rootdir,
                                   one_channel_alpha_trigger_cfg, 'insert', master_random_state_object)

    ############# Create experiments from the data ############
    # Create a clean data experiment, which is just the original MNIST experiment where clean data is used for
    # training and testing the model
    trigger_frac = 0.0
    trigger_behavior = tdb.WrappedAdd(1, 10)
    e = tde.ClassicExperiment(toplevel_folder, trigger_behavior)
    train_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean', 'train_mnist.csv'),
                                   clean_dataset_rootdir,
                                   mod_filename_filter='*train*',
                                   split_clean_trigger=False,
                                   trigger_frac=trigger_frac)
    train_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_train.csv'), index=None)
    test_clean_df, test_triggered_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean',
                                                                        'test_mnist.csv'),
                                                           clean_dataset_rootdir,
                                                           mod_filename_filter='*test*',
                                                           split_clean_trigger=True,
                                                           trigger_frac=trigger_frac)
    test_clean_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_test_clean.csv'), index=None)
    test_triggered_df.to_csv(os.path.join(toplevel_folder, 'mnist_clean_experiment_test_triggered.csv'), index=None)

    # Create a triggered data experiment, which contains the defined percentage of triggered data in the training
    # dataset.  The remaining training data is clean data.  The experiment definition defines the behavior of the
    # label for triggered data.  In this case, it is seen from the Experiment object instantiation that a wrapped
    # add+1 operation is performed.
    # In the code below, we create several experiments with varying levels of poisoned data to allow for
    # experimentation.
    trigger_fracs = [0.05, 0.1, 0.15, 0.2]
    for trigger_frac in trigger_fracs:
        train_df = e.create_experiment(os.path.join(toplevel_folder, 'mnist_clean', 'train_mnist.csv'),
                                       os.path.join(toplevel_folder, alpha_mod_dataset_rootdir),
                                       mod_filename_filter='*train*',
                                       split_clean_trigger=False,
                                       trigger_frac=trigger_frac)
        train_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                     '_experiment_train.csv'), index=None)
        test_clean_df, test_triggered_df = e.create_experiment(os.path.join(toplevel_folder,
                                                                            'mnist_clean', 'test_mnist.csv'),
                                                               os.path.join(toplevel_folder, alpha_mod_dataset_rootdir),
                                                               mod_filename_filter='*test*',
                                                               split_clean_trigger=True,
                                                               trigger_frac=trigger_frac)
        test_clean_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                          '_experiment_test_clean.csv'), index=None)
        test_triggered_df.to_csv(os.path.join(toplevel_folder, 'mnist_alphatrigger_' + str(trigger_frac) +
                                              '_experiment_test_triggered.csv'), index=None)



from typing import Sequence, Union
import logging
import os
import math
import json
import subprocess
import copy
import tempfile

from .config import UGEConfig

from .model_generator_interface import ModelGeneratorInterface, validate_model_generator_interface_input
from .config import ModelGeneratorConfig

logger = logging.getLogger(__name__)
ALL_EXEC_PERMISSIONS = 0o555

"""
This file contains all the functionality needed to train models for a Univa Grid Engine (UGE) HPC cluster.
"""


class UGEModelGenerator(ModelGeneratorInterface):
    """
    Class which generates models utilizing a Univa Grid Engine
    """
    def __init__(self, configs: Union[ModelGeneratorConfig, Sequence[ModelGeneratorConfig]],
                 uge_config: UGEConfig, working_directory: str = os.path.join(os.environ['HOME'],"uge_model_generator"),
                 validate_uge_dirs: bool = True):
        """
        Initializes a UGE Model Generator
        :param configs: a ModelGeneratorConfig or a Sequence of ModelGeneratorConfig objects which
                        define the models to be created
        :param uge_config: configuration object which specifies how the parallelization should be farmed across the UGE.
        :param working_directory: the directory where the scripts to be run will be stored, along with any objects
            that are persisted to carry out the parallelization.
            NOTE: this should be a directory that is replicated across the cluster (for example, sometimes /tmp is
             configured to not replicate across the cluster
        :param validate_uge_dirs: if True, the directory will be validated to ensure it doesn't begin with /tmp,
            for the reason that typically /tmp is not replicated across the cluster.
        """
        super().__init__(configs)
        self.uge_config = uge_config
        if self.uge_config.multi_model_same_gpu:
            self.configs_expanded = self.configs
        else:
            self.configs_expanded = self.expand_modelgen_configs_to_process()

        self.working_directory = working_directory
        self.validate_uge_dirs = validate_uge_dirs
        self.validate()

    def expand_modelgen_configs_to_process(self) -> Sequence[ModelGeneratorConfig]:
        """
        Converts a sequence of ModelGeneratorConfig objects into another sequence of ModelGeneratorConfig
        objects such that each element in the sequence only creates one model.
        For example:
          Input: cfgs = [cfg1->num_models=1, cfg2->num_models=2].  len(cfgs)=2
          Output: cfgs = [cfg1->num_models=1, cfg2->num_models=1, cfg2->num_models=1]. len(cfgs)=3
        This is useful so that we can fully distribute all the models that need to be generated
        ########################################
        NOTE: This will lead to multiple configs pointing to the same data on disk.  I'm not sure if
            this is a problem for PyTorch or not, but this is something to investigate if unexpected results arise.
        ########################################
        :return: expanded config configuration
        """
        configs_expanded = []
        for cfg in self.configs:
            num_models = cfg.num_models
            for model_idx in range(num_models):
                # make a copy of the object so that there are no memory conflicts among multiple jobs running
                cfg_copy = copy.deepcopy(cfg)
                cfg_copy.num_models = 1
                configs_expanded.append(cfg_copy)
        return configs_expanded

    def get_queue_numjobs_assignment(self) -> Sequence:
        """
        Determine the number of jobs to give to each queue based on UGEConfig
        :return: a list of tuples, with each tuple containing the queue in index-0, and the number of jobs
                 assigned to that queue in index-1
        """
        num_available_queues = len(self.uge_config.queues)
        num_jobs_to_process = len(self.configs_expanded)
        num_jobs_to_assign = num_jobs_to_process
        queue_numjobs_assignment = []
        if self.uge_config.queue_distribution is None:
            num_jobs = math.ceil(num_jobs_to_process/num_available_queues)
            for q in self.uge_config.queues:
                if num_jobs_to_assign < num_jobs:
                    num_jobs = num_jobs_to_assign
                queue_numjobs_assignment.append((q, num_jobs))
                num_jobs_to_assign -= num_jobs
                if num_jobs_to_assign <= 0:
                    break
        else:
            for ii, q in enumerate(self.uge_config.queues):
                desired_dist_value = self.uge_config.queue_distribution[ii]
                num_jobs = num_jobs_to_assign*desired_dist_value
                if num_jobs_to_assign < num_jobs:
                    num_jobs = num_jobs_to_assign
                queue_numjobs_assignment.append((q, num_jobs))
                num_jobs_to_assign -= num_jobs
                if num_jobs_to_assign <= 0:
                    break

        return queue_numjobs_assignment

    @staticmethod
    def _gen_py_script(pyscript_fname: str, pyscript_log_fname: str, modelgen_cfg_persist_fname: str,
                      persist_metadata_fname: str,
                      run_id: str = None, filename: str = None) -> None:
        """
        Generate the Python script which will be used to
        :param pyscript_fname: name of the file which will have the Python script
        :param pyscript_log_fname: log filename where all Python program output will be captured
        :param modelgen_cfg_persist_fname: filename of where the configuration data will be persisted for
                distributing the job onto cluster nodes
        :param persist_metadata_fname: filename of where the configuration metadata will be persisted for
                distributing the job onto cluster nodes
        :param run_id: any specified run-id which will be passed to the Runner
        :param filename: any specified filename which will be passed to the runner
        :return: None
        """
        with open(pyscript_fname, 'w') as f:
            f.write('''\
#!/usr/bin/env python
import json
import logging.config
import trojai.modelgen.config as tpmc
import trojai.modelgen.runner as tpmr

# setup logger
logging.config.dictConfig({
    'version': 1,
    'formatters': {
        'detailed': {
            'format': '[%%(asctime)s] %%(levelname)s in %%(module)s: %%(message)s',
        },
    },
    'handlers': {
        'file': {
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': '%s',
            'maxBytes': 10 * 1024 * 1024,
            'backupCount': 5,
            'formatter': 'detailed',
            'level': 'INFO',
        },
    },
    'loggers': {
        'trojai': {
            'handlers': ['file'],
        },
        'trojai_private': {
            'handlers': ['file'],
        },
    },
    'root': {
        'level': 'INFO',
    },
})

modelgen_cfg = tpmc.ModelGeneratorConfig.load("%s")
with open("%s", 'r') as f:
    persist_metadata = json.load(f)
run_cfg = tpmc.modelgen_cfg_to_runner_cfg(modelgen_cfg, run_id=%s, filename=%s)

runner = tpmr.Runner(run_cfg, persist_metadata=persist_metadata, progress_bar_disable=True)
runner.run()
            ''' % (pyscript_log_fname, modelgen_cfg_persist_fname, persist_metadata_fname, run_id, filename))

    @staticmethod
    def _gen_bash_script(bashscript_fname: str, pyscript_fname: str) -> None:
        """
        Generates the bash script, which sets up the correct environment for each node and calls the generated Python
        script to generate the models
        :param bashscript_fname: the filename of the bash script to be generated
        :param pyscript_fname: the filename of the python script which will be called in the bash script
        :return: None
        """
        pyscript_abs_path = os.path.abspath(pyscript_fname)
        with open(bashscript_fname, 'w') as f:
            f.write('''\
#!/bin/bash
source /etc/profile.d/modules.sh
module load cuda91
# setup conda environment
. /cm/shared/apps/anaconda3/etc/profile.d/conda.sh
conda activate trojai

python3 %s
            ''' % (pyscript_abs_path,))
        os.chmod(bashscript_fname, ALL_EXEC_PERMISSIONS)  # give everyone read & execute permissions

    @staticmethod
    def _gen_bash_command(bashscript_fname: str, uge_log_fname: str, queue_name: str,
                         gpu_node: bool = False, sync_mode: bool = False) -> str:
        """
        Creates a UGE command to submit the job to the cluster for processing
        :param bashscript_fname: the bash script to run by the cluster
        :param uge_log_fname: a filename of any log messages captured by the UGE
        :param queue_name: the queue to submit the job to
        :param gpu_node: (bool) indicates whether the queue the job is being submitted to has GPU nodes
        :param sync_mode: if True, then the shell will be captured by this process.  Currently unsupported!
        :return: (str) bash command
        """
        # UGE job submit command example
        #qsub -q gpu-k40.q -l gpu=1 -V -v PATH -cwd -S /bin/bash -j y -sync y -o /home/karrak1/trojai/qsub.log
        #test_torch_cuda.sh

        cmd_list = list()
        cmd_list.append('qsub -q')
        cmd_list.append(queue_name)
        if gpu_node:
            cmd_list.append('-l gpu=1')
        cmd_list.append('-V -v PATH -cwd -S /bin/bash -j y')
        if sync_mode:
            cmd_list.append('-sync y')
        cmd_list.append('-o %s' % (uge_log_fname,))  # logging
        cmd_list.append(bashscript_fname)
        cmd = ' '.join(cmd_list)
        return cmd

    def run(self, mock=False) -> None:
        """
        Run's the actual UGE job.
        :param mock: if True, then it generates all the necessary scripts but doesn't execute the UGE command
        :return: None
        """
        modelgen_cfgs_processed_idx = 0
        queue_numjobs_assignment = self.get_queue_numjobs_assignment()
        for qj_assignment in queue_numjobs_assignment:
            queue = qj_assignment[0]
            num_jobs = qj_assignment[1]
            queue_subworking_dir_name = queue.queue_name
            for job_idx in range(num_jobs):
                # setup working directory for this job
                job_working_dir = str(job_idx)

                modelgen_cfg_to_schedule = self.configs_expanded[modelgen_cfgs_processed_idx]
                num_models_to_gen = modelgen_cfg_to_schedule.num_models
                for model_idx in range(num_models_to_gen):
                    subjob_working_dir = os.path.join(self.working_directory, queue_subworking_dir_name,
                                                      job_working_dir, str(model_idx))

                    try:
                        os.makedirs(subjob_working_dir)
                    except IOError as e:
                        logger.exception(e)
                        raise IOError(e)

                    filename = None
                    run_id = None
                    if modelgen_cfg_to_schedule.filenames is not None:
                        if isinstance(modelgen_cfg_to_schedule.filenames, str):
                            filename = modelgen_cfg_to_schedule.filenames
                        else:
                            filename = modelgen_cfg_to_schedule.filenames[model_idx]
                    elif modelgen_cfg_to_schedule.run_ids is not None:
                        run_id = modelgen_cfg_to_schedule.run_ids[model_idx]

                    # save a serialized version of the modelgen_config object to the working directory
                    persist_cfg_fname = os.path.join(subjob_working_dir, 'persist_config')
                    modelgen_cfg_to_schedule.save(persist_cfg_fname)
                    # save the experiment information as a persist_metadata to track results after
                    persist_metadata_fname = os.path.join(subjob_working_dir, 'persist_metadata.json')
                    persist_metadata = modelgen_cfg_to_schedule.experiment_cfg
                    with open(persist_metadata_fname, 'w') as f:
                        json.dump(persist_metadata, f)

                    # setup filenames for the python & bash scripts which will be generated to distribute the processing
                    pyscript_fname = os.path.join(subjob_working_dir, "generate_model.py")
                    bashscript_fname = os.path.join(subjob_working_dir, "generate_model_" +
                                                                        queue.queue_name +
                                                                        "_" + str(job_idx) + "_" + str(model_idx) +
                                                                        ".sh")
                    pyscript_log_fname = os.path.join(subjob_working_dir, "generate_model.py.log")
                    uge_log_fname = os.path.join(subjob_working_dir, "log.txt")

                    # create the python script that will run the actual model
                    UGEModelGenerator._gen_py_script(pyscript_fname, pyscript_log_fname,
                                                     persist_cfg_fname,
                                                     persist_metadata_fname,
                                                     run_id, filename)

                    # create the bash script wrapper that will be passed to UGE
                    UGEModelGenerator._gen_bash_script(bashscript_fname, pyscript_fname)

                    # create the command that will be called to submit the job
                    bash_cmd = UGEModelGenerator._gen_bash_command(bashscript_fname, uge_log_fname, queue.queue_name,
                                                                   queue.gpu_enabled, queue.sync_mode)

                    # submit the job
                    logger.info("submitting job with command: " + bash_cmd)
                    if not mock:
                        try:
                            subprocess.run(bash_cmd, shell=True, check=True)
                        except subprocess.CalledProcessError as e:
                            logger.exception(e)
                            raise subprocess.CalledProcessError(e)

                modelgen_cfgs_processed_idx += 1

    def validate(self) -> None:
        """
        Validate the input configuration
        """
        validate_model_generator_interface_input(self.configs)
        if not isinstance(self.uge_config, UGEConfig):
            msg = "uge_queue_config must be of type UGEQueueConfig"
            logger.error(msg)
            raise TypeError(msg)

        os_temp_dir = tempfile.gettempdir()

        if not isinstance(self.working_directory, str):
            msg = "working_directory must be a path to a directory that the UGEModelGenerator can use to submit jobs"
            logger.error(msg)
            raise TypeError(msg)
        else:
            # check if the working directory is in /tmp, which is not replicated across the cluster and thus should
            # not be used
            if self.validate_uge_dirs:
                if self.working_directory.startswith(os_temp_dir):
                    msg = os_temp_dir + " should not be used for the working directory because OS temp directories " \
                                        "are typically not propagated throughout the cluster!"
                    logger.error(msg)
                    raise ValueError(msg)
            try:
                os.makedirs(self.working_directory)
            except IOError as e:
                logger.exception(e)

        if not isinstance(self.validate_uge_dirs, bool):
            msg = "validate_uge_dirs must be a boolean!"
            logger.error(msg)
            raise TypeError(msg)

        if self.validate_uge_dirs:
            for cfg in self.configs_expanded:
                if cfg.model_save_dir.startswith(os_temp_dir):
                    msg = os_temp_dir + " should not be used as the directory for saving models because OS temp " \
                                        "directories are typically not propagated throughout the cluster!"
                    logger.error(msg)
                    raise ValueError(msg)
                if cfg.stats_save_dir.startswith(os_temp_dir):
                    msg = os_temp_dir + " should not be used as the directory for saving stats because OS temp " \
                                        "directories are typically not propagated throughout the cluster!"
                    logger.error(msg)
                    raise ValueError(msg)



import logging

from .label_behavior import LabelBehavior

logger = logging.getLogger(__name__)

"""
Defines some common behaviors which are used to modify labels when designing an experiment with triggered and clean data
"""


class WrappedAdd(LabelBehavior):
    """
    Adds a defined amount to each input label, with an optional maximum value around which labels are wrapped
    """
    def __init__(self, add_val: int, max_num_classes: int = None) -> None:
        """
        Creates the WrappedAdd object
        :param add_val: the value to add to each input label
        :param max_num_classes: the maximum number of classes such that modified labels are wrapped
        """
        self.add_val = add_val
        self.max_num_classes = max_num_classes

    def do(self, y_true: int) -> int:
        """
        Performs the actual specified label modification
        :param y_true: input label to be modified
        :return: the modified label
        """
        modified_label = y_true + self.add_val
        if self.max_num_classes is not None:
            modified_label %= self.max_num_classes
        logger.info("Converted label %d to %d" % (y_true, modified_label))
        return modified_label



import logging

logger = logging.getLogger(__name__)

"""
Configurations for various types of data
"""


class DataConfiguration:
    pass


class TextDataConfiguration(DataConfiguration):
    def __init__(self, max_vocab_size: int = 25000,
                 embedding_dim: int = 100, embedding_type: str = 'glove',
                 num_tokens_embedding_train: str = '6B'):
        """

        :param max_vocab_size: integer indicating maximum vocabulary size
        :param embedding_dim: valid options are: [50, 100, 200, 300]
        :param embedding_type: valid options are: ['glove']
        :param num_tokens_embedding_train: NOTE: only used if embedding_dim is 300, otherwise ignored!
        """
        self.max_vocab_size = max_vocab_size
        self.embedding_dim = embedding_dim
        self.embedding_type = embedding_type
        self.embedding_vectors_cfg = None
        self.num_tokens_embedding_train = num_tokens_embedding_train

        self.validate()
        self.set_embedding_vectors_cfg()

    def set_embedding_vectors_cfg(self):
        if self.embedding_dim == 50:
            self.embedding_vectors_cfg = 'glove.6B.50d'
        elif self.embedding_dim == 100:
            self.embedding_vectors_cfg = 'glove.6B.100d'
        elif self.embedding_dim == 200:
            self.embedding_vectors_cfg = 'glove.6B.200d'
        elif self.embedding_dim == 300:
            if self.num_tokens_embedding_train == '6B':
                self.embedding_vectors_cfg = 'glove.6B.300d'
            elif self.num_tokens_embedding_train == '42B':
                self.embedding_vectors_cfg = 'glove.42B.300d'
            elif self.num_tokens_embedding_train == '840B':
                self.embedding_vectors_cfg = 'glove.840B.300d'

    def validate(self):
        if not isinstance(self.max_vocab_size, int):
            msg = "max_vocab_size must be an integer"
            logger.error(msg)
            raise ValueError(msg)
        if self.max_vocab_size < 1:
            msg = "max_vocab_size must be atleast 1!"
            logger.error(msg)
            raise ValueError(msg)
        if not isinstance(self.embedding_dim, int):
            msg = "embedding_dim must be an integer"
            logger.error(msg)
            raise ValueError(msg)
        if self.embedding_dim < 1:
            msg = "embedding dimension must be atleast 1"
            logger.error(msg)
            raise ValueError(msg)
        if self.embedding_type not in ['glove']:
            msg = "embedding type must be one of: ['glove']"
            logger.error(msg)
            raise ValueError(msg)
        if self.embedding_dim not in [50, 100, 200, 300]:
            msg = "embedding dimension must be one of [50, 100, 200, 300]"
            logger.error(msg)
            raise ValueError(msg)
        if self.num_tokens_embedding_train not in ['6B', '42B', '840B']:
            msg = "number of tokens to train the embedding must be one of: ['6B', '42B', '840B']"
            logger.error(msg)
            raise ValueError(msg)


class ImageDataConfiguration(DataConfiguration):
    pass


import collections
import logging
import os
from typing import Sequence
import copy
import cloudpickle as pickle

import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
from torch.utils.data import Dataset
from torch.utils.tensorboard import SummaryWriter
from tqdm import tqdm

from .training_statistics import BatchStatistics, EpochStatistics
from .optimizer_interface import OptimizerInterface
from .config import DefaultOptimizerConfig
from .constants import VALID_OPTIMIZERS

logger = logging.getLogger(__name__)


def _eval_acc(y_hat: torch.Tensor, y_truth: torch.Tensor, n_total: int = 0, n_correct: int = 0):
    """
    Wrapper for computing accuracy
    :param y_hat: the computed predictions, should be of shape (n_batches, num_classes)
    :param y_truth: the actual y-values
    :param n_total: the total number of data points processed, this will be incremented and returned
    :param n_correct: the total number of correct predictions so far, before this function was called
    :return: accuracy, updated n_total, updated n_correct

    TODO:
     [ ] - need to handle the case where the user applies sigmoid at the output of the final layer before
           outputting.  With the current behavior, _eval_acc would apply the sigmoid function twice
     [ ] - are there non-sigmoid conversions we'd want to support when rounding predictions w/ one output class?
    """
    y_hat_size = y_hat.size()
    if len(y_hat_size) == 2:
        n_batches, num_classes = y_hat.size()
    elif len(y_hat_size) == 1:
        n_batches = y_hat_size[0]
        num_classes = 1
    else:
        msg = "unhandled size of y_hat!:" + str(y_hat.size())
        logger.error(msg)
        raise ValueError(msg)
    n_total += n_batches
    if num_classes > 1:
        max_index = y_hat.max(dim=1)[1]
        n_correct += (max_index == y_truth).sum().item()
    else:
        rounded_preds = torch.round(torch.sigmoid(y_hat))
        n_correct += (rounded_preds.int() == y_truth.int()).sum().item()
    acc = 100. * float(n_correct) / float(n_total)
    return acc, n_total, n_correct


def train_val_dataset_split(dataset: torch.utils.data.Dataset, split_amt: float) \
        -> (torch.utils.data.Dataset, torch.utils.data.Dataset):
    """
    Splits a PyTorch dataset (of type: torch.utils.data.Dataset) into train/test
    TODO:
      [ ] - specify random seed to torch splitter
    :param dataset: the dataset to be split
    :param split_amt: fraction specificing the validation dataset size relative to the whole.  1-split_amt will
                      be the size of the training dataset
    :return: a tuple of the train and validation datasets
    """

    if split_amt < 0 or split_amt > 1:
        msg = "Dataset split amount must be between 0 and 1"
        logger.error(msg)
        raise ValueError(msg)

    dataset_len = len(dataset)
    train_len = int(dataset_len * (1 - split_amt))
    val_len = int(dataset_len - train_len)
    lengths = [train_len, val_len]
    train_dataset, val_dataset = torch.utils.data.random_split(dataset, lengths)
    return train_dataset, val_dataset


class DefaultOptimizer(OptimizerInterface):
    """
    Defines the default optimizer which trains the models
    """

    def __init__(self, optimizer_cfg: DefaultOptimizerConfig = None):
        """
        Initializes the default optimizer with a DefaultOptimizerConfig
        :param optimizer_cfg: the configuration used to initialize the DefaultOptimizer
        """
        if optimizer_cfg is None:
            logger.info("Using default parameters to setup Optimizer!")
            self.optimizer_cfg = DefaultOptimizerConfig()
        elif not isinstance(optimizer_cfg, DefaultOptimizerConfig):
            msg = "optimizer_cfg must be of type DefaultOptimizerConfig"
            logger.error(msg)
            raise TypeError(msg)
        else:
            self.optimizer_cfg = optimizer_cfg

        # setup parameters for training here
        self.device = self.optimizer_cfg.training_cfg.device

        self.loss_function_str = self.optimizer_cfg.training_cfg.objective.lower()
        if self.loss_function_str == "cross_entropy_loss".lower():
            self.loss_function = nn.CrossEntropyLoss()
        elif self.loss_function_str == 'BCEWithLogitsLoss'.lower():
            self.loss_function = nn.BCEWithLogitsLoss()
        self.loss_function.to(self.device)

        self.lr = self.optimizer_cfg.training_cfg.lr
        self.optimizer_str = self.optimizer_cfg.training_cfg.optim.lower()
        self.optimizer = None

        self.batch_size = self.optimizer_cfg.training_cfg.batch_size
        self.num_epochs = self.optimizer_cfg.training_cfg.epochs
        self.save_best_model = self.optimizer_cfg.training_cfg.save_best_model

        self.str_description = "{'batch_size':%d, 'num_epochs':%d, 'device':'%s', 'lr':%.5e, 'loss_function':'%s', " \
                               "'optimizer':'%s'}" % \
                               (self.batch_size, self.num_epochs, self.device.type, self.lr, self.loss_function_str,
                                self.optimizer_str)

        # setup configuration for logging and tensorboard
        self.num_batches_per_logmsg = self.optimizer_cfg.reporting_cfg.num_batches_per_logmsg
        self.num_epochs_per_metrics = self.optimizer_cfg.reporting_cfg.num_epochs_per_metrics
        self.num_batches_per_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_metrics
        # NOTE: the configuration parameter 'num_batches_per_val_dataset_metrics' has the most significant impact on
        #  training performance.  It can be useful for debugging development of a model, but when scaling up,
        #  set this parameter to None to disable computing the validation dataset and speed up training
        self.num_batches_per_val_dataset_metrics = self.optimizer_cfg.reporting_cfg.num_batches_per_val_dataset_metrics
        if self.device.type == 'cpu' and self.num_batches_per_val_dataset_metrics is not None:
            logger.warning('Training will be VERY SLOW on a CPU with num_batches_per_val_dataset_metrics set to a '
                           'value other than None.  If validation dataset metrics are still desired, '
                           'consider increasing this value to speed up training')

        tensorboard_output_dir = self.optimizer_cfg.reporting_cfg.tensorboard_output_dir
        self.tb_writer = None
        if tensorboard_output_dir is not None:
            self.tb_writer = SummaryWriter(tensorboard_output_dir)

        optimizer_cfg_str = 'Optimizer[%s] Configured as: loss[%s], learning-rate[%.5e], batch-size[%d] ' \
                            'num-epochs[%d] Device[%s]' % \
                            (self.optimizer_str, str(self.loss_function), self.lr, self.batch_size, self.num_epochs,
                             self.device.type)
        reporting_cfg_str = 'Reporting Configured as: num_batches_per_log_message[%d] tensorboard_dir[%s]' % \
                            (self.num_batches_per_logmsg, tensorboard_output_dir)
        metrics_capture_str = 'Metrics capturing configured as: num_epochs_per_metric[%d] ' \
                              'num_batches_per_epoch_per_metric[%d]' % \
                              (self.num_epochs_per_metrics, self.num_batches_per_metrics)
        logger.info(self.str_description)
        logger.info(optimizer_cfg_str)
        logger.info(reporting_cfg_str)
        logger.info(metrics_capture_str)

    def __str__(self) -> str:
        return self.str_description

    def __deepcopy__(self, memodict={}):
        optimizer_cfg_copy = copy.deepcopy(self.optimizer_cfg)
        # WARNING: this assumes that none of the derived attributes have been changed after construction!
        return DefaultOptimizer(DefaultOptimizerConfig(optimizer_cfg_copy.training_cfg,
                                                       optimizer_cfg_copy.reporting_cfg))

    def get_cfg_as_dict(self) -> dict:
        output_dict = dict(device=str(self.device.type),
                           epochs=self.num_epochs,
                           batch_size=self.batch_size,
                           learning_rate=self.lr,
                           optim=self.optimizer_str,
                           objective=self.loss_function_str)
        return output_dict

    def __eq__(self, other) -> bool:
        try:
            if self.optimizer_cfg == other.optimizer_cfg:
                # we still check the derived attributes to ensure that they remained the same after
                # after construction
                if self.device.type == other.device.type and self.loss_function_str == other.loss_function_str and \
                        self.lr == other.lr and self.optimizer_str == other.optimizer_str and \
                        self.batch_size == other.batch_size and self.num_epochs == other.num_epochs and \
                        self.str_description == other.str_description and \
                        self.num_batches_per_logmsg == other.num_batches_per_logmsg and \
                        self.num_epochs_per_metrics == other.num_epochs_per_metrics and \
                        self.num_batches_per_metrics == other.num_batches_per_metrics and \
                        self.num_batches_per_val_dataset_metrics == other.num_batches_per_val_dataset_metrics:
                    if self.tb_writer is not None:
                        if other.tb_writer is not None:
                            if self.tb_writer.log_dir == other.tb_writer.log_dir:
                                return True
                            else:
                                return False
                        else:
                            return False
                    else:
                        if other.tb_writer is not None:
                            return False
                        else:
                            # both are None
                            return True
            else:
                return False
        except AttributeError:
            return False

    def get_device_type(self) -> str:
        """
        :return: a string representing the device used to train the model
        """
        return self.device.type

    def save(self, fname: str) -> None:
        """
        Saves the configuration object used to construct the DefaultOptimizer.
        NOTE: because the DefaultOptimizer object itself is not persisted, but rather the
          DefaultOptimizerConfig object, the state of the object is not persisted!
        :param fname: the filename to save the DefaultOptimizer's configuration.
        :return: None
        """
        self.optimizer_cfg.save(fname)

    @staticmethod
    def load(fname: str) -> OptimizerInterface:
        """
        Reconstructs a DefaultOptimizer, by loading the configuration used to construct the original
        DefaultOptimizer, and then creating a new DefaultOptimizer object from the saved configuration
        :param fname: The filename of the saved optimzier
        :return: a DefaultOptimizer object
        """
        with open(fname, 'rb') as f:
            loaded_optimzier_cfg = pickle.load(f)
        return DefaultOptimizer(loaded_optimzier_cfg)

    def _eval_loss_function(self, y_hat: torch.Tensor, y_truth: torch.Tensor) -> torch.Tensor:
        """
        Wrapper for evaluating the loss function to abstract out any data casting we need to do
        :param y_hat: the predicted y-value
        :param y_truth: the actual y-value
        :return: the loss associated w/ the prediction and actual
        """
        if self.loss_function_str == "cross_entropy_loss":
            train_loss = self.loss_function(y_hat, y_truth.long())
        else:
            train_loss = self.loss_function(y_hat, y_truth)
        return train_loss

    def train(self, net: torch.nn.Module, dataset: torch.utils.data.Dataset, train_val_split: float = 0.0,
              progress_bar_disable: bool = False, torch_dataloader_kwargs: dict = None) \
        -> (torch.nn.Module, Sequence[EpochStatistics]):
        """
        Train the network.
        :param net: the network to train
        :param dataset: the dataset to train the network on
        :param train_val_split: the % of training data to use as validation data
        :param progress_bar_disable: if True, disables the progress bar
        :param torch_dataloader_kwargs: any additional kwargs to pass to PyTorch's native DataLoader
        :return: the trained network, and a list of EpochStatistics objects which contain the statistics for training
        """
        net = net.to(self.device)

        net.train()  # put network into training mode
        if self.optimizer_str == 'adam':
            self.optimizer = optim.Adam(net.parameters(), lr=self.lr)
        elif self.optimizer_str not in VALID_OPTIMIZERS:
            msg = self.optimizer_str + " is not a supported optimizer!"
            logger.error(msg)
            raise ValueError(msg)
        else:
            msg = self.optimizer_str + " not yet implemented!"
            logger.error(msg)
            raise NotImplementedError(msg)

        # set according to the following guidelines:
        # https://discuss.pytorch.org/t/when-to-set-pin-memory-to-true/19723
        # https://pytorch.org/docs/stable/notes/cuda.html
        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True

        # split into train & validation datasets, and setup data loaders
        data_loader_kwargs_in = {} if torch_dataloader_kwargs is None else torch_dataloader_kwargs
        logger.info('DataLoader[Train/Val] kwargs=' + str(torch_dataloader_kwargs))

        train_dataset, val_dataset = train_val_dataset_split(dataset, train_val_split)
        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        train_loader = DataLoader(train_dataset, batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True,
                                  **data_loader_kwargs_in)
        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        val_loader = DataLoader(val_dataset, batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True,
                                **data_loader_kwargs_in)

        # use validation in training? provide as option?
        all_epochs_stats = []
        best_net = None
        best_validation_acc = -999
        best_training_acc = -999
        for epoch_idx, epoch in enumerate(range(self.num_epochs)):
            compute_batch_stats = True if epoch_idx % self.num_epochs_per_metrics == 0 else False
            batches_stats = self.train_epoch(net, train_loader, val_loader, epoch, compute_batch_stats,
                                             progress_bar_disable=progress_bar_disable)

            if compute_batch_stats and len(batches_stats) > 0:
                epoch_training_stats = EpochStatistics(epoch_idx)
                epoch_training_stats.add_batch(batches_stats)
                all_epochs_stats.append(epoch_training_stats)

                if self.save_best_model:
                    if train_val_split == 0.0:
                        # use training accuracy as the metric for deciding the best model
                        final_batch_training_acc = batches_stats[-1].batch_train_accuracy
                        if final_batch_training_acc >= best_training_acc:
                            msg = "Updating best model with epoch:[%d] accuracy[%0.02f].  Previous best training " \
                                  "accuracy was: %0.02f" % (epoch_idx, final_batch_training_acc, best_training_acc)
                            logger.info(msg)
                            best_net = net
                            best_training_acc = final_batch_training_acc
                    else:
                        # use validation accuracy as the metric for deciding the best model
                        final_batch_validation_acc = batches_stats[-1].batch_validation_accuracy
                        if final_batch_validation_acc >= best_validation_acc:
                            msg = "Updating best model with epoch:[%d] accuracy[%0.02f].  Previous best validation " \
                                  "accuracy was: %0.02f" % (epoch_idx, final_batch_validation_acc, best_validation_acc)
                            logger.info(msg)
                            best_net = net
                            best_validation_acc = final_batch_validation_acc

        if self.save_best_model:
            return best_net, all_epochs_stats
        else:
            return net, all_epochs_stats

    def train_epoch(self, model: nn.Module, train_loader: DataLoader, val_loader: DataLoader,
                    epoch_num: int, compute_batch_stats: bool = True,
                    avg_loss_num_batches: int = 100, progress_bar_disable: bool = False):
        """
        Runs one epoch of training on the specified model
        TODO:
          [ ] - pass in the avg_loss_num_batches parameter cleanly and configurably

        :param model: the model to train for one epoch
        :param train_loader: a DataLoader object pointing to the training dataset
        :param val_loader: a DataLoader object pointing to the validation dataset
        :param epoch_num: the epoch number that is being trained
        :param compute_batch_stats: if True, computes statistics for the batch based on the reporting configuration
                specified in the initialization of the optimizer
        :param avg_loss_num_batches: the number of batches of data to accumulate to compute average loss
        :param progress_bar_disable: if True, disables the progress bar
        :return: a list of statistics for batches where statistics were computed
        """
        loop = tqdm(train_loader, disable=progress_bar_disable)

        pid = os.getpid()
        train_dataset_len = len(train_loader.dataset)
        train_loader_len = len(train_loader)

        # NOTE: potential speed-up by not computing the average ... but this seems like premature optimization to me
        avg_train_loss_circbuf = collections.deque(maxlen=avg_loss_num_batches)
        avg_val_loss_vec = np.empty(len(val_loader.dataset))

        train_n_correct, train_n_total = 0, 0
        val_n_correct, val_n_total = 0, 0
        val_acc, val_loss = None, None
        batch_stats = []
        for batch_idx, (x, y_truth) in enumerate(loop):
            x = x.to(self.device)
            y_truth = y_truth.to(self.device)

            # put network into training mode & zero out previous gradient computations
            model.train()
            self.optimizer.zero_grad()

            # get predictions based on input & weights learned so far
            y_hat = model(x)

            # compute metrics
            batch_train_loss = self._eval_loss_function(y_hat, y_truth)
            running_train_acc, train_n_total, train_n_correct = _eval_acc(y_hat, y_truth,
                                                                          n_total=train_n_total,
                                                                          n_correct=train_n_correct)

            # compute gradient
            batch_train_loss.backward()
            self.optimizer.step()

            if len(val_loader) > 0 and self.num_batches_per_val_dataset_metrics is not None and \
                    ((batch_idx % self.num_batches_per_val_dataset_metrics == 0) or
                     (batch_idx % self.num_batches_per_metrics == 0)):
                # last condition ensures metrics are computed for storage put model into evaluation mode
                model.eval()
                # turn off auto-grad for validation set computation
                with torch.no_grad():
                    for val_batch_idx, (x_eval, y_truth_eval) in enumerate(val_loader):
                        x_eval = x_eval.to(self.device)
                        y_truth_eval = y_truth_eval.to(self.device)
                        y_hat_eval = model(x_eval)

                        val_loss_tensor = self._eval_loss_function(y_hat_eval, y_truth_eval)
                        val_loss = val_loss_tensor.item()
                        val_acc, val_n_total, val_n_correct = _eval_acc(y_hat_eval, y_truth_eval,
                                                                        n_total=val_n_total,
                                                                        n_correct=val_n_correct)
                        avg_val_loss_vec[val_batch_idx] = val_loss

                avg_val_loss = np.mean(avg_val_loss_vec)
                if self.tb_writer is not None:
                    try:
                        self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name +
                                                  '-avg_validation_loss', avg_val_loss)
                    except:
                        # TODO: catch specific expcetions
                        pass

            # NOTE: should these be options to compute training accuracy, etc, as well?
            avg_train_loss_circbuf.append(batch_train_loss.item())
            avg_train_loss = np.mean(avg_train_loss_circbuf)

            loop.set_description('Epoch {}/{}'.format(epoch_num + 1, self.num_epochs))
            loop.set_postfix(avg_train_loss=avg_train_loss)

            # report batch statistics to tensorflow
            if self.tb_writer is not None:
                try:
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-train_loss',
                                              batch_train_loss.item())
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-avg_train_loss',
                                              avg_train_loss)
                    self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-running_train_acc',
                                              running_train_acc)
                    if len(val_loader) > 0 and self.num_batches_per_val_dataset_metrics is not None:
                        self.tb_writer.add_scalar(self.optimizer_cfg.reporting_cfg.experiment_name + '-val_acc', val_acc)
                except:
                    # TODO: catch specific expcetions
                    pass

            # save batch statistics
            if compute_batch_stats and (batch_idx % self.num_batches_per_metrics == 0):
                batch_stat = BatchStatistics(batch_idx, running_train_acc, batch_train_loss.item(),
                                             val_acc, val_loss)
                batch_stats.append(batch_stat)

            if batch_idx % self.num_batches_per_logmsg == 0:
                logger.info('{}\tTrain Epoch: {} [{}/{} ({:.0f}%)]\tAvgTrainLoss: {:.6f}\tTrainAcc: {:.6f}'.format(
                    pid, epoch_num, batch_idx * len(x), train_dataset_len,
                                    100. * batch_idx / train_loader_len, avg_train_loss, running_train_acc))

        return batch_stats

    def test(self, net: nn.Module, clean_data: Dataset, triggered_data: Dataset,
             progress_bar_disable: bool = False, torch_dataloader_kwargs: dict = None) -> dict:
        """
        Test the trained network
        :param net: the trained module to run the test data through
        :param clean_data: the clean Dataset
        :param triggered_data: the triggered Dataset, if None, not computed
        :param progress_bar_disable: if True, disables the progress bar
        :param torch_dataloader_kwargs: any keyword arguments to pass directly to PyTorch's DataLoader
        :return: a dictionary of the statistics on the clean and triggered data (if applicable)
        """
        test_data_statistics = {}
        net.eval()
        data_loader_kwargs_in = {} if torch_dataloader_kwargs is None else torch_dataloader_kwargs
        logger.info('DataLoader[Test] kwargs=' + str(torch_dataloader_kwargs))

        pin_memory = False
        if self.device.type != 'cpu':
            pin_memory = True
        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        data_loader = DataLoader(clean_data, batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True,
                                 **data_loader_kwargs_in)

        # test type is classification accuracy on clean and triggered data
        test_n_correct = 0
        test_n_total = 0
        with torch.no_grad():
            for batch, (x, y_truth) in enumerate(data_loader):
                x = x.to(self.device)
                y_truth = y_truth.to(self.device)
                y_hat = net(x)
                test_acc, test_n_total, test_n_correct = _eval_acc(y_hat, y_truth,
                                                                   n_total=test_n_total,
                                                                   n_correct=test_n_correct)
        test_data_statistics['clean_accuracy'] = test_acc
        test_data_statistics['clean_n_total'] = test_n_total
        logger.info("Accuracy on clean test data: %0.02f" %
                    (test_data_statistics['clean_accuracy'],))

        if triggered_data is None:
            return test_data_statistics

        # drop_last=True is from: https://stackoverflow.com/questions/56576716
        data_loader = DataLoader(triggered_data, batch_size=self.batch_size, pin_memory=pin_memory, drop_last=True)
        test_n_correct = 0
        test_n_total = 0
        with torch.no_grad():
            for batch, (x, y_truth) in enumerate(data_loader):
                x = x.to(self.device)
                y_truth = y_truth.to(self.device)
                y_hat = net(x)
                test_acc, test_n_total, test_n_correct = _eval_acc(y_hat, y_truth,
                                                                   n_total=test_n_total,
                                                                   n_correct=test_n_correct)
        test_data_statistics['triggered_accuracy'] = test_acc
        test_data_statistics['triggered_n_total'] = test_n_total
        logger.info("Accuracy on triggered test data: %0.02f" %
                    (test_data_statistics['triggered_accuracy'],))
        return test_data_statistics



import csv
import os
import shutil
from typing import Tuple, Sequence, Dict

import cv2
import mnist_utils
import numpy as np
import pandas as pd
from numpy.random import RandomState
from tqdm import tqdm

import trojai.datagen.constants as dg_constants
import trojai.datagen.image_entity as dg_entity
import trojai.datagen.transform_interface as dg_transform
import trojai.datagen.utils as dg_utils

"""
Module containing the necessary functions to create clean MNIST data.
"""


def load_dataset(csv_path: str) -> Tuple:
    """
    Load MNIST data in CSV format into memory for processing
    :param csv_path: the path to the CSV file containing the MNIST data
    :return: a tuple of the data and the output labels
    """
    df = pd.read_csv(csv_path, header=None)
    output_labels = df[0].values
    data = df[range(1, df.shape[1])].values

    return data, output_labels


def _df_iterate_store(df_X: pd.DataFrame, df_y: pd.Series, fname_prefix: str, rootdir: str, subdir: str,
                      xforms: Sequence[dg_transform.Transform], random_state_obj: RandomState,
                      output_file_start_counter: int = 0,
                      dtype=np.uint8) -> Sequence[Dict]:
    """
    Helper function which iterates over a dataframe that contains the MNIST data, applies the defined transformations
    to the data, and creates & stores an output for each row in the input dataframe in the specified folder structure.
    :param df_X: A dataframe containing the MNIST data, where each row contains a flattend matrix of the MNIST digit
                 representation
    :param df_y: A series containing the label; the indices of df_X and df_y must be synced
                 i.e. the label specified in df_y[i] must correspond to the data in df_X[i]
    :param fname_prefix: filename prefix of the output data
    :param rootdir: root directory into which the data will be stored
    :param subdir: the sub directory into which the data will be stored
    :param xforms: a list of transforms to be applied to each image before it is stored
    :param random_state_obj: object used to derive random states for each image that is generated
    :param output_file_start_counter: output files have the format: <fname_prefix>_counter, and this value denotes the
           start value of that counter
    :param dtype: how to interpret the input data from df_X
    :return: a list of dictionaries of the paths to the files that were stored as a result of the processing, and their
            associated label.
    """
    output_list = []
    for ii in tqdm(range(df_X.shape[0]), desc=fname_prefix):
        # setup the random state for the image from the master
        img_random_state = RandomState(random_state_obj.randint(dg_constants.RANDOM_STATE_DRAW_LIMIT))

        output_fname = os.path.join(subdir, fname_prefix+'_'+str(output_file_start_counter)+'.png')
        output_filename_fullpath = os.path.join(rootdir, output_fname)
        output_file_start_counter += 1

        X = df_X[ii, :].reshape(mnist_utils.MNIST_IMG_SHAPE).astype(dtype)
        # add a new axis to make the image compatible w/ tensors (nrows, ncols, nchan)
        X = X[:, :, np.newaxis]
        X_obj = dg_entity.GenericImageEntity(X, mask=None)

        # perform any "pre" modifications to the data if specified
        X_obj = dg_utils.process_xform_list(X_obj, xforms, img_random_state)
        y = df_y[ii]

        # write data to disk
        cv2.imwrite(output_filename_fullpath, X_obj.get_data())

        # append to list
        output_list.append({'file': output_fname, 'label': y})

    return output_list


def _validate_create_clean_dataset_cfgdict(xform_list):
    for xform in xform_list:
        if not isinstance(xform, dg_transform.Transform):
            return False

    return True


def create_clean_dataset(input_train_csv_path: str, input_test_csv_path: str,
                         output_rootdir: str, output_train_csv_file: str, output_test_csv_file: str,
                         train_fname_prefix: str, test_fname_prefix: str, xforms: Sequence[dg_transform.Transform],
                         random_state_obj: RandomState) -> None:
    """
    Creates a "clean" MNIST dataset, which is a the MNIST dataset (with potential transformations applied),
    but no triggers.
    :param input_train_csv_path: path to the CSV file containing the training data.  The format of the CSV file is
                                 is specified by the mnist_utils.convert() function
    :param input_test_csv_path:  path to the CSV file containing the test data.  The format of the CSV file is
                                 is specified by the mnist_utils.convert() function
    :param output_rootdir: the root directory into which the clean data will be stored.
                            training data will be stored in: output_rootdir/train
                            test data will be stored in: output_rootdir/test
    :param output_train_csv_file: a CSV file of the training data, which specifies paths to files, and their
                                  associated labels
    :param output_test_csv_file: a CSV file of the test data, which specifies paths to files, and their
                                  associated labels
    :param train_fname_prefix: a prefix to every training filename
    :param test_fname_prefix: a prefix to every test filename
    :param xforms: a dictionary which contains the necessary transformations to be applied to each input image.
                    The configuration is validated by _validate_create_clean_dataset_cfgdict(), but at a high level,
                    the dictionary must contain the 'transforms' key, and that must be a list of transformations to
                    be applied.
    :param random_state_obj: object used to derive random states for each image that is generated
    :return: None
    """
    # input error checking
    if not _validate_create_clean_dataset_cfgdict(xforms):
        raise ValueError("mod_cfg argument incorrectly specified!")

    # create a fresh version of the directory
    try:
        shutil.rmtree(output_rootdir)
    except:
        pass

    X_train, y_train = load_dataset(input_train_csv_path)
    X_test, y_test = load_dataset(input_test_csv_path)
    train_output_subdir = 'train'
    test_output_subdir = 'test'

    # make necessary sub-directories
    try:
        os.makedirs(os.path.join(output_rootdir, train_output_subdir))
    except:
        pass
    try:
        os.makedirs(os.path.join(output_rootdir, test_output_subdir))
    except:
        pass

    random_state = random_state_obj.get_state()
    clean_train_output_list = _df_iterate_store(X_train, y_train,
                                                train_fname_prefix, output_rootdir,
                                                train_output_subdir,
                                                xforms,
                                                random_state_obj,
                                                output_file_start_counter=0)
    # reset state to ensure reproducibility regardless of the # of data points generated
    random_state_obj.set_state(random_state)
    clean_test_output_list = _df_iterate_store(X_test, y_test,
                                               test_fname_prefix, output_rootdir,
                                               test_output_subdir,
                                               xforms,
                                               random_state_obj,
                                               output_file_start_counter=0)

    keys = ['file', 'label']
    with open(os.path.join(output_rootdir, output_train_csv_file), 'w') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(clean_train_output_list)
    with open(os.path.join(output_rootdir, output_test_csv_file), 'w') as output_file:
        dict_writer = csv.DictWriter(output_file, keys)
        dict_writer.writeheader()
        dict_writer.writerows(clean_test_output_list)



import json
import os
import logging
import types
import glob
import uuid
import time

import torch
import torch.nn as nn

from .config import RunnerConfig, DefaultOptimizerConfig
from .training_statistics import TrainingRunStatistics
from .default_optimizer import DefaultOptimizer
from .optimizer_interface import OptimizerInterface

logger = logging.getLogger(__name__)


def add_numerical_extension(path, filename):
    # check if any files already exist in that directory w/ digit extensions or not, and get the filename of interest
    existing_fnames = glob.glob(os.path.join(path, filename+'.*'))
    if len(existing_fnames) > 0:
        # remove the .json & csv files from consideration
        existing_fnames = [os.path.basename(x) for x in existing_fnames if '.json' not in x]
        existing_fnames = [os.path.basename(x) for x in existing_fnames if '.csv' not in x]
        max_cur_digit_ext = 1
        max_cur_digit_fname_without_ext, _ = os.path.splitext(existing_fnames[0])
        # iterate through the filenames and find the maximum integer extension
        for f in existing_fnames:
            fname_without_ext, ext = os.path.splitext(f)
            try:
                ext_val = int(ext[1:])  # the [1:] is needed to remove the . from the extension
                if ext_val > max_cur_digit_ext:
                    max_cur_digit_ext = ext_val
                    max_cur_digit_fname_without_ext = fname_without_ext
            except ValueError:
                pass
        next_digit_ext = max_cur_digit_ext + 1
        fname_to_return = max_cur_digit_fname_without_ext + '.' + (str(next_digit_ext))
    else:
        fname_without_ext, ext = os.path.splitext(filename)
        try:
            cur_digit_ext = int(ext[1:])  # the [1:] is needed to remove the . from the extension
            next_digit_ext = cur_digit_ext + 1
            fname_to_return = fname_without_ext+'.'+(str(next_digit_ext))
        except ValueError:
            fname_to_return = filename+'.1'

    return fname_to_return


class Runner:
    """
    Fundamental unit of model generation, which trains a model as specified in a RunnerConfig object.
    """
    def __init__(self, runner_cfg: RunnerConfig,
                 persist_metadata: dict = None,
                 progress_bar_disable: bool = False):
        """
        Initialize a model runner, which sets up the Optimizer, passes data to the optimizer, and collects the
        trained model and associated statistics
        :param runner_cfg: (RunnerConfig) Object that contains necessary data and objects to train a model using
            this runner.
        :param persist_metadata: (dict), if not None, the contents of this are appended to the output summary
        dictionary. This can allow for easy tracking of results if they are being collated by an additional process.
        """
        if not isinstance(runner_cfg, RunnerConfig):
            msg = "Expected a RunnerConfig object for argument 'runner_config', instead got " \
                            "type {}".format(type(runner_cfg))
            logger.error(msg)
            raise TypeError(msg)
        self.cfg = runner_cfg
        # todo: make this a type check like with runner_cfg? To reduce confusion if metadata is not a dict but code
        #   runs; make warning
        if persist_metadata is None or not isinstance(persist_metadata, dict):
            msg = "Argument 'persist_metadata' was not None nor type 'dict'. Argument will be ignored."
            logger.warning(msg)
            self.persist_info = {}
        else:
            self.persist_info = persist_metadata

        self.progress_bar_disable = progress_bar_disable

    def run(self) -> None:
        """Trains a model and saves it and the associated model statistics"""
        train_data, clean_test_data, triggered_test_data, \
            train_dataset_desc, clean_test_dataset_desc, triggered_test_dataset_desc = self.cfg.data.load_data()
        arch_factory_kwargs = {} if self.cfg.arch_factory_kwargs is None else self.cfg.arch_factory_kwargs
        torch_dataloader_kwargs = {} if self.cfg.data.torch_dataloader_kwargs is None else \
            self.cfg.data.torch_dataloader_kwargs

        if self.cfg.arch_factory_kwargs_generator is not None:
            arch_factory_kwargs.update(self.cfg.arch_factory_kwargs_generator(train_dataset_desc,
                                                                              clean_test_dataset_desc,
                                                                              triggered_test_dataset_desc))

        model = self.cfg.arch_factory.new_architecture(**arch_factory_kwargs)
        if self.cfg.parallel:
            num_available_gpus = torch.cuda.device_count()
            logger.info("Attempting to use " + str(num_available_gpus) + " GPUs for training!")
            model = nn.DataParallel(model)

        model_stats = TrainingRunStatistics()
        # TODO: this is hacked to deal w/ text data, we need to make this better
        training_cfg_list = []
        t1 = time.time()
        if isinstance(train_data, types.GeneratorType):
            for data, optimizer in zip(train_data, self.cfg.optimizer_generator):  # both are generators
                model, epoch_training_stats = optimizer.train(model, data, self.cfg.train_val_split,
                                                              self.progress_bar_disable, torch_dataloader_kwargs)
                model_stats.add_epoch(epoch_training_stats)
                # add training configuration information to data to be saved
                training_cfg_list.append(self._get_training_cfg(optimizer))
        else:
            optimizer = next(self.cfg.optimizer_generator)
            model, epoch_training_stats = optimizer.train(model, train_data, self.cfg.train_val_split,
                                                          self.progress_bar_disable, torch_dataloader_kwargs)
            model_stats.add_epoch(epoch_training_stats)
            # add training configuration information to data to be saved
            training_cfg_list.append(self._get_training_cfg(optimizer))
        t2 = time.time()
        # NOTE: The test function used here is one corresponding to the last optimizer used for training. An exception
        #  will be raised if no training occurred, but validation code prior to this line should prevent this from
        #  ever happening.
        test_acc = optimizer.test(model, clean_test_data, triggered_test_data, self.progress_bar_disable,
                                  torch_dataloader_kwargs)
        t3 = time.time()

        # Save model train/test statistics and other relevant information
        model_stats.autopopulate_final_summary_stats()
        model_stats.set_final_clean_data_test_acc(test_acc['clean_accuracy'])
        model_stats.set_final_clean_data_n_total(test_acc['clean_n_total'])
        model_stats.set_final_triggered_data_test_acc(test_acc.get('triggered_accuracy', None))
        model_stats.set_final_triggered_data_n_total(test_acc.get('triggered_n_total', None))

        # add training/test wall-times to stats
        self.persist_info['training_wall_time_sec'] = t2-t1
        self.persist_info['test_wall_time_sec'] = t3-t2

        self._save_model_and_stats(model, model_stats, training_cfg_list)

    @staticmethod
    def _get_training_cfg(optimizer):
        if isinstance(optimizer, DefaultOptimizerConfig):
            training_cfg = optimizer.training_cfg.get_cfg_as_dict()
        elif isinstance(optimizer, DefaultOptimizer) or isinstance(optimizer, OptimizerInterface):
            training_cfg = optimizer.get_cfg_as_dict()
        else:
            msg = "Unable to get training_cfg from optimizer(_cfg): {}, returning empty dict".format(optimizer)
            logger.warning(msg)
            training_cfg = dict()
        return training_cfg

    def _save_model_and_stats(self, model: nn.Module, stats: TrainingRunStatistics, training_cfg_list: list):
        model_path = self.cfg.model_save_dir
        if not os.path.isdir(model_path):
            try:
                os.makedirs(model_path)
            except IOError as e:
                logger.error(e)
        stats_path = self.cfg.stats_save_dir
        if not os.path.isdir(model_path):
            try:
                os.makedirs(model_path)
            except IOError as e:
                logger.error(e)

        extn = '.pt'

        if self.cfg.filename is not None:
            filename = self.cfg.filename
            if os.path.splitext(filename)[1] != extn:
                filename += extn
        else:
            if self.cfg.run_id is not None:
                filename = model.__class__.__name__ + '_id' + str(self.cfg.run_id)
            else:
                filename = model.__class__.__name__
            if self.persist_info is not None and 'name' in self.persist_info:
                filename += '_' + self.persist_info['name']
            filename += extn

        if self.cfg.save_with_hash:
            filename += '.' + str(uuid.uuid1().hex)
        else:
            filename = add_numerical_extension(model_path, filename)

        model.eval()
        model_output_fname = os.path.join(model_path, filename)
        stats_output_fname = os.path.join(stats_path, filename+'.stats.json')
        detailed_stats_output_fname = os.path.join(stats_path, filename+'.stats.detailed.csv')

        logger.info("Saving trained model to " + str(model_output_fname) + " in PyTorch format.")
        if self.cfg.parallel:
            model = model.module
        # TODO: should we move the model to a CPU before saving, to prevent GPU memory spike when loading?
        torch.save(model, model_output_fname)
        model_training_stats_dict = stats.get_summary()
        for i, cfg in enumerate(training_cfg_list):
            model_training_stats_dict.update({"optimizer_"+str(i): cfg})
        # add experiment configuration to the dictionary which gets printed
        model_training_stats_dict.update(self.persist_info)

        # send the statistics to the logger
        logger.info(str(model_training_stats_dict))

        # save the entire dict as a json object
        with open(stats_output_fname, 'w') as fp:
            json.dump(model_training_stats_dict, fp)
        # save detailed statistics
        stats.save_detailed_stats_to_disk(detailed_stats_output_fname)



import logging

import cv2
import numpy as np
from numpy.random import RandomState

from .image_entity import ImageEntity, GenericImageEntity
from .transform_interface import Transform

logger = logging.getLogger(__name__)

"""
Module contains various classes that relate to size transformations of input objects
"""


class Resize(Transform):
    """
    Resizes an Entity
    """
    def __init__(self, new_size: tuple = (200, 200), interpolation: int = cv2.INTER_CUBIC) -> None:
        """
        Initialize the resizer object
        :param new_size: a tuple of the size in pixes for x and y dimensions
        :param interpolation: the interpolation method to resize the input Entity
        """
        self.new_size = new_size
        self.interpolation = interpolation

    def do(self, img_obj: ImageEntity, random_state_obj: RandomState) -> ImageEntity:
        """
        Perform the resizing
        :param img_obj: The input object to be resized according the specified configuration
        :param random_state_obj: ignored
        :return: The resized object
        """
        img_out = cv2.resize(img_obj.get_data(), self.new_size, interpolation=self.interpolation)
        mask_out = cv2.resize(img_obj.get_mask().astype(np.float32), self.new_size,
                              interpolation=self.interpolation).astype(bool)
        logger.info("Resized image of shape=%s to shape=%s using %d interpolation" %
                    (str(img_obj.get_data().shape), str(self.new_size), self.interpolation))
        return GenericImageEntity(img_out, mask_out)



RANDOM_STATE_DRAW_LIMIT = 2**32 - 1
'''
In the data generation process, every new entity that is generated gets a new random seed by drawing from 
np.random.RandomState.randint(), where the RandomState object comes from a master RandomState created at the 
beginning of the data generation process.  The constant RANDOM_STATE_DRAW_LIMIT defines the argument passed into the 
randint(...) call.

The reason we create a new seed for every Entity is to enable reproducibility.  Each Entity that is created may go 
through a series of transformations that include randomness at various stages.  As such, having a seed associated 
with each Entity will enable us to reproduce those specific random variations easily.
'''



import collections.abc
import json
import logging
from typing import Union, Sequence
import csv

logger = logging.getLogger(__name__)

"""
Contains classes necessary for collecting statistics on the model during training
"""


class BatchStatistics:
    """
    Represents the statistics collected from processing a batch
    """
    def __init__(self, batch_num: int,
                 batch_train_accuracy: float,
                 batch_train_loss: float,
                 batch_validation_accuracy: float,
                 batch_validation_loss: float):
        """
        :param batch_num: (int) batch number of collected statistics
        :param batch_train_accuracy: (float) training set accuracy for this batch
        :param batch_train_loss: (float) training loss for this batch
        :param batch_validation_accuracy: (float) validation set accuracy for this batch
        :param batch_validation_loss: (float) validation set loss for this batch
        """
        self.batch_num = batch_num
        self.batch_train_accuracy = batch_train_accuracy
        self.batch_train_loss = batch_train_loss
        self.batch_validation_accuracy = batch_validation_accuracy
        self.batch_validation_loss = batch_validation_loss

    def get_batch_num(self):
        return self.batch_num

    def get_batch_train_acc(self):
        return self.batch_train_accuracy

    def get_batch_train_loss(self):
        return self.batch_train_loss

    def get_batch_validation_acc(self):
        return self.batch_validation_accuracy

    def get_batch_validation_loss(self):
        return self.batch_validation_loss

    def set_batch_train_acc(self, acc):
        if 0 <= acc <= 100:
            self.batch_train_accuracy = acc
        else:
            msg = "Batch training accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_batch_train_loss(self, loss):
        self.batch_train_loss = loss

    def set_batch_validation_acc(self, acc):
        if acc is None or 0 <= acc <= 100:  # allow for None in case validation metrics are NOT computed for efficiency
            self.batch_validation_accuracy = acc
        else:
            msg = "Batch validation accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_batch_validation_loss(self, loss):
        self.batch_validation_loss = loss


class EpochStatistics:
    """
    Contains the statistics computed for an Epoch
    """
    def __init__(self, epoch_num):
        self.epoch_num = epoch_num
        self.batch_stats_list = []

    def add_batch(self, batch_stats: Union[BatchStatistics, Sequence[BatchStatistics]]):
        if isinstance(batch_stats, collections.abc.Sequence):
            self.batch_stats_list.extend(batch_stats)
        else:
            self.batch_stats_list.append(batch_stats)

    def get_epoch_num(self):
        return self.epoch_num

    def get_batch_stats(self):
        return self.batch_stats_list


class TrainingRunStatistics:
    """
    Contains the statistics computed for an entire training run
    TODO:
     [ ] - have another function which returns detailed statistics per epoch in an easily serialized manner
    """
    def __init__(self):
        self.epoch_stats_list = []

        self.final_train_acc = 0.
        self.final_train_loss = 0.
        self.final_val_acc = 0.
        self.final_val_loss = 0.
        self.final_clean_data_test_acc = 0.
        self.final_clean_data_n_total = 0
        self.final_triggered_data_test_acc = None
        self.final_triggered_data_n_total = None

    def add_epoch(self, epoch_stats: Union[EpochStatistics, Sequence[EpochStatistics]]):
        if isinstance(epoch_stats, collections.abc.Sequence):
            self.epoch_stats_list.extend(epoch_stats)
        else:
            self.epoch_stats_list.append(epoch_stats)

    def get_epochs_stats(self):
        return self.epoch_stats_list

    def autopopulate_final_summary_stats(self):
        """
        Uses the information from the final epoch's final batch to auto-populate the following statistics:
            final_train_acc
            final_train_loss
            final_val_acc
            final_val_loss
        """
        final_epoch_stats = self.epoch_stats_list[-1]
        final_batch_stats = final_epoch_stats.get_batch_stats()[-1]
        self.set_final_train_acc(final_batch_stats.get_batch_train_acc())
        self.set_final_train_loss(final_batch_stats.get_batch_train_loss())
        self.set_final_val_acc(final_batch_stats.get_batch_validation_acc())
        self.set_final_val_loss(final_batch_stats.get_batch_validation_loss())

    def set_final_train_acc(self, acc):
        if 0 <= acc <= 100:
            self.final_train_acc = acc
        else:
            msg = "Final Training accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_train_loss(self, loss):
        self.final_train_loss = loss

    def set_final_val_acc(self, acc):
        if acc is None or 0 <= acc <= 100:  # allow for None in case validation metrics are not computed
            self.final_val_acc = acc
        else:
            msg = "Final validation accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_val_loss(self, loss):
        self.final_val_loss = loss

    def set_final_clean_data_test_acc(self, acc):
        if 0 <= acc <= 100:
            self.final_clean_data_test_acc = acc
        else:
            msg = "Final clean data test accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_triggered_data_test_acc(self, acc):
        # we allow None in order to indicate that triggered data wasn't present in this dataset
        if acc is None or 0 <= acc <= 100:
            self.final_triggered_data_test_acc = acc
        else:
            msg = "Final triggered data test accuracy should be between 0 and 100!"
            logger.error(msg)
            raise ValueError(msg)

    def set_final_clean_data_n_total(self, n):
        self.final_clean_data_n_total = n

    def set_final_triggered_data_n_total(self, n):
        if n is None or n > 0:
            self.final_triggered_data_n_total = n
        else:
            msg = "Triggered dataset size must be > 0!"
            logger.error(msg)
            raise ValueError(msg)

    def get_summary(self):
        """
        Returns a dictionary of the summary statistics from the training run
        """
        summary_dict = dict()
        summary_dict['final_train_acc'] = self.final_train_acc
        summary_dict['final_train_loss'] = self.final_train_loss
        summary_dict['final_val_acc'] = self.final_val_acc
        summary_dict['final_val_loss'] = self.final_val_loss
        summary_dict['final_clean_data_test_acc'] = self.final_clean_data_test_acc
        summary_dict['final_triggered_data_test_acc'] = self.final_triggered_data_test_acc
        summary_dict['final_clean_data_n_total'] = self.final_clean_data_n_total
        summary_dict['final_triggered_data_n_total'] = self.final_triggered_data_n_total

        return summary_dict

    def save_summary_to_json(self, json_fname: str) -> None:
        """
        Saves the training summary to a JSON file
        """
        summary_dict = self.get_summary()
        # write it to json
        with open(json_fname, 'w') as fp:
            json.dump(summary_dict, fp)
        logger.info("Wrote summary statistics: %s to %s" % (str(summary_dict), json_fname))

    def save_detailed_stats_to_disk(self, fname: str) -> None:
        """
        Saves all batch statistics for every epoch as a CSV file
        :param fname: filename to save the detailed information to
        :return: None
        """
        output_data = []
        for e in self.epoch_stats_list:
            batches_stats = e.get_batch_stats()
            for batch_num, batch_stats in enumerate(batches_stats):
                row = dict(epoch_number=e.epoch_num,
                           batch_num=batch_num,
                           train_accuracy=batch_stats.get_batch_train_acc(),
                           train_loss=batch_stats.get_batch_train_loss(),
                           val_acc=batch_stats.get_batch_validation_acc(),
                           val_loss=batch_stats.get_batch_validation_loss())
                output_data.append(row)
        # write it as a csv
        if len(output_data) > 0:
            keys = output_data[0].keys()
            with open(fname, 'w') as output_file:
                dict_writer = csv.DictWriter(output_file, keys)
                dict_writer.writeheader()
                dict_writer.writerows(output_data)
            logger.info("Wrote detailed statistics to %s" % (fname, ))



import copy
import logging
import os
from typing import Callable, Any, Union, Sequence
import types

import pandas as pd
from torch.utils.data import Dataset

from .constants import VALID_DATA_TYPES
from .datasets import CSVDataset, CSVTextDataset
from .data_descriptions import DataDescription
from .data_configuration import DataConfiguration

logger = logging.getLogger(__name__)


class DataManager:
    """ Manages data from an experiment from trogai.datagen. """
    def __init__(self, experiment_path: str, train_file: Union[str, Sequence[str]], clean_test_file: str,
                 triggered_test_file: str = None,
                 data_type: str = 'image',
                 data_transform: Callable[[Any], Any] = (lambda x: x),
                 label_transform: Callable[[int], int] = lambda y: y,
                 file_loader: Union[Callable[[str], Any], str] = 'default_image_loader',
                 shuffle_train=True, shuffle_clean_test=False, shuffle_triggered_test=False,
                 data_configuration: DataConfiguration = None,
                 custom_datasets: dict = None,
                 torch_dataloader_kwargs: dict = None):
        """
        Initializes the DataManager object
        :param experiment_path: (str) absolute path to experiment data.
        :param train_file: (str) csv file name(s) of the training data. If iterable is provided, all will be trained
            on before model will be tested
        :param clean_test_file: (str) csv file name of the clean test data.
        :param triggered_test_file: (str) csv file name of the triggered test data.
        :param data_type: (str) can be 'image', 'text', or 'custom'.  The TrojaiDataManager uses this to determine how
                          to load the actual data and prepare it to be fed into the optimizer.
        :param data_transform: (function: any -> any) how to transform the data (e.g. and image) to fit into the
            desired model and objective function; optional
            NOTE: Currently - this argument is only used if data_type='image'
        :param label_transform: (function: int->int) how to transform the label to the data; optional
            NOTE: Currently - this argument is only used if data_type='image'
        :param file_loader: (function: str->any or str) how to create the data object to pass into an architecture
            from a file path, or default loader to use. Options include: 'default_image_loader'
            default: 'default_image_loader'
            NOTE: Currently - this argument is only used if data_type='image'
        :param shuffle_train: (bool) shuffle the training data before training; default=True
        :param shuffle_clean_test: (bool) shuffle the clean test data; default=False
        :param shuffle_triggered_test (bool) shuffle the triggered test data; default=False
        :param data_configuration - a DataConfiguration object that might be useful for setting up
                how data is loaded
        :param custom_datasets - if data_type is 'custom', then the custom_datasets is a user implementation of
                torch.utils.data.Dataset.  We expect a dictionary of datasets, where the expected dictionary will
                look as follows:
                    {
                        'train': Union[torch.utils.data.Dataset, Sequence[torch.utils.data.Dataset]],
                        'clean_test': torch.utils.data.Dataset,
                        'triggered_test': Union[torch.utils.data.Dataset, None]
                    }
        """

        self.experiment_path = experiment_path
        try:
            iter(train_file)
        except TypeError:
            pass
        if type(train_file) == str:
            train_file = [train_file]
        self.train_file = train_file
        self.clean_test_file = clean_test_file
        self.triggered_test_file = triggered_test_file

        self.data_type = data_type
        self.data_loader = file_loader
        self.data_transform = data_transform
        self.label_transform = label_transform

        self.shuffle_train = shuffle_train
        self.shuffle_clean_test = shuffle_clean_test
        self.shuffle_triggered_test = shuffle_triggered_test

        self.data_configuration = data_configuration
        self.datasets = custom_datasets
        self.torch_dataloader_kwargs = torch_dataloader_kwargs

        self.validate()

    def __deepcopy__(self, memodict={}):
        return DataManager(self.experiment_path, self.train_file, self.clean_test_file,
                           self.triggered_test_file, self.data_type, copy.deepcopy(self.data_transform),
                           copy.deepcopy(self.label_transform), copy.deepcopy(self.data_loader),
                           self.shuffle_train, self.shuffle_clean_test, self.shuffle_triggered_test,
                           self.data_configuration, self.datasets, self.torch_dataloader_kwargs)

    def __eq__(self, other):
        if self.experiment_path == other.experiment_path and self.train_file == other.train_file and \
           self.clean_test_file == other.clean_test_file and self.triggered_test_file == other.triggered_test_file and \
           self.data_type == other.data_type and \
           self.data_transform == other.data_transform and self.label_transform == other.label_transform and \
           self.data_loader == other.data_loader and self.shuffle_train == other.shuffle_train and \
           self.shuffle_clean_test == other.shuffle_clean_test and \
           self.shuffle_triggered_test == other.shuffle_triggered_test and \
           self.data_configuration == other.data_configuration and \
           self.torch_dataloader_kwargs == other.torch_dataloader_kwargs:
            # Note: when we compare callables, we simply compare whether the callable is the same reference in memory
            #  or not.  This means that if two callables are functionally equivalent, but are different object
            #  references then the equality comparison will fail
            #  This note also pertains to comparing the dataset_obj, which is a torch.utils.data.Dataset object!

            # TODO: compare datasets

            return True
        else:
            return False

    def load_data(self):
        """
        Load experiment data as given from initialization.
        :return: Objects containing training and test, and triggered data if it was provided.

        TODO:
         [ ] - extend the text data-type to have more input arguments, for example the tokenizer and FIELD options
         [ ] - need to support sequential training for text datasets
        """
        if self.data_type == 'image':
            logger.info("Loading Training Dataset")
            first_dataset = CSVDataset(self.experiment_path, self.train_file[0],
                                       data_transform=self.data_transform,
                                       label_transform=self.label_transform,
                                       data_loader=self.data_loader,
                                       shuffle=self.shuffle_train)
            train_dataset = (first_dataset if ii == 0 else CSVDataset(self.experiment_path, self.train_file[ii],
                                                                      data_transform=self.data_transform,
                                                                      label_transform=self.label_transform,
                                                                      data_loader=self.data_loader,
                                                                      shuffle=self.shuffle_train)
                             for ii in range(len(self.train_file)))

            if self.clean_test_file is not None:
                clean_test_dataset = CSVDataset(self.experiment_path, self.clean_test_file,
                                                data_transform=self.data_transform,
                                                label_transform=self.label_transform,
                                                data_loader=self.data_loader,
                                                shuffle=self.shuffle_clean_test)
                if len(clean_test_dataset) == 0:
                    clean_test_dataset = None
                    msg = 'Clean Test Dataset was empty and will be skipped...'
                    logger.info(msg)
            else:
                clean_test_dataset = None
                msg = 'Clean Test Dataset was empty and will be skipped...'
                logger.info(msg)
            if self.triggered_test_file is not None:
                triggered_test_dataset = CSVDataset(self.experiment_path, self.triggered_test_file,
                                                    data_transform=self.data_transform,
                                                    label_transform=self.label_transform,
                                                    data_loader=self.data_loader,
                                                    shuffle=self.shuffle_triggered_test)
                if len(triggered_test_dataset) == 0:
                    triggered_test_dataset = None
                    msg = 'Triggered Dataset was empty, testing on triggered data will be skipped...'
                    logger.info(msg)
            else:
                triggered_test_dataset = None
                msg = 'Triggered Dataset was empty, testing on triggered data will be skipped...'
                logger.info(msg)

            # nothing to fill in at the moment for image, we can update as needed
            if isinstance(train_dataset, types.GeneratorType):
                train_dataset_desc = first_dataset.get_data_description()
            else:
                train_dataset_desc = train_dataset.get_data_description()
            if clean_test_dataset is not None:
                clean_test_dataset_desc = clean_test_dataset.get_data_description()
            else:
                clean_test_dataset_desc = None
            if triggered_test_dataset is not None:
                triggered_test_dataset_desc = triggered_test_dataset.get_data_description()
            else:
                triggered_test_dataset_desc = None

        elif self.data_type == 'text':
            if len(self.train_file) > 1:
                msg = "Sequential Training not supported for Text datatype!"
                logger.error(msg)
                raise ValueError(msg)
            # ensure a DataDescription is set for text data!
            if self.data_configuration is None:
                msg = "data_configuration object needs to be set for Text data processing!"
                logger.error(msg)
                raise ValueError(msg)

            logger.info("Loading Training Dataset")
            train_dataset = CSVTextDataset(self.experiment_path, self.train_file[0], shuffle=self.shuffle_train)
            train_dataset.build_vocab(self.data_configuration.embedding_vectors_cfg,
                                      self.data_configuration.max_vocab_size)
            # pass in the learned vocabulary from the training data to the clean test dataset

            clean_test_dataset = None
            if self.clean_test_file is not None:
                clean_test_dataset = CSVTextDataset(self.experiment_path, self.clean_test_file,
                                                    text_field=train_dataset.text_field,
                                                    label_field=train_dataset.label_field,
                                                    shuffle=self.shuffle_clean_test)
                if len(clean_test_dataset) == 0:
                    msg = 'Clean Test Dataset was empty and will be skipped...'
                    logger.info(msg)
            else:
                msg = 'Clean Test Dataset was empty and will be skipped...'
                logger.info(msg)
            triggered_test_dataset = None
            if self.triggered_test_file is not None:
                logger.info("Loading Triggered Test Dataset")
                # pass in the learned vocabulary from the training data to the triggered test dataset
                triggered_test_dataset = CSVTextDataset(self.experiment_path, self.triggered_test_file,
                                                        text_field=train_dataset.text_field,
                                                        label_field=train_dataset.label_field,
                                                        shuffle=self.shuffle_triggered_test)
                if len(triggered_test_dataset) == 0:
                    msg = 'Triggered Dataset was empty, testing on triggered data will be skipped...'
                    logger.info(msg)
                    triggered_test_dataset = None
            else:
                triggered_test_dataset = None

            train_dataset_desc = train_dataset.get_data_description()
            if clean_test_dataset is not None and len(clean_test_dataset) > 0:
                clean_test_dataset_desc = clean_test_dataset.get_data_description()
            else:
                clean_test_dataset_desc = None
            if triggered_test_dataset is not None and len(triggered_test_dataset) > 0:
                triggered_test_dataset_desc = triggered_test_dataset.get_data_description()
            else:
                triggered_test_dataset_desc = None
        elif self.data_type == 'custom':
            train_dataset = self.datasets['train']
            clean_test_dataset = self.datasets['clean_test']
            # using the "get" function to get elements from dictionary ensures that we return None if the keys were
            # not provided
            triggered_test_dataset = self.datasets.get('triggered_test')
            if train_dataset is not None:
                train_dataset_desc = train_dataset.get_data_description()
            else:
                train_dataset_desc = None
            if clean_test_dataset is not None:
                clean_test_dataset_desc = clean_test_dataset.get_data_description()
            else:
                clean_test_dataset_desc = None
            if triggered_test_dataset is not None:
                triggered_test_dataset_desc = triggered_test_dataset.get_data_description()
            else:
                triggered_test_dataset_desc = None
        else:
            msg = "Unsupported data_type argument provided"
            logger.error(msg)
            raise NotImplementedError(msg)

        return train_dataset, clean_test_dataset, triggered_test_dataset, \
            train_dataset_desc, clean_test_dataset_desc, triggered_test_dataset_desc

    def validate(self) -> None:
        """
        Validate the construction of the TrojaiDataManager object
        :return: None

        TODO:
         [ ] - think about whether the contents of the files passed into the DataManager should be validated,
               in addition to simply checking for existence, which is what is done now
        """
        if self.torch_dataloader_kwargs is not None and not isinstance(self.torch_dataloader_kwargs, dict):
            msg = "torch_dataloader_kwargs must be a dictionary!"
            logger.error(msg)
            raise ValueError(msg)

        if self.data_type == 'custom':
            if self.datasets is None:
                msg = "dataset_obj must not be None if data_type is set to Custom"
                logger.error(msg)
                raise ValueError(msg)
            elif isinstance(self.datasets, dict):
                required_keys_to_test = ['train', 'clean_test']
                optional_keys_to_test = ['triggered_test']
                dd_keys_to_test = ['train_data_description', 'clean_test_data_description',
                                   'triggered_test_data_description']
                for k in required_keys_to_test:
                    if k in self.datasets:
                        if isinstance(self.datasets[k], Dataset):
                            pass
                        else:
                            msg = "The expected type of value for key:" + k + ' is Dataset'
                            logger.error(msg)
                            raise ValueError(msg)
                    else:
                        msg = 'Expected key:' + k + ' in datasets dictionary'
                        logger.error(msg)
                        raise ValueError(msg)
                for k in optional_keys_to_test:
                    if k in self.datasets:
                        if isinstance(self.datasets[k], Dataset):
                            pass
                        else:
                            msg = "The expected type of value for key:" + k + ' is Dataset'
                            logger.error(msg)
                            raise ValueError(msg)
                for k in dd_keys_to_test:
                    if k in self.datasets:
                        if self.datasets[k] is None or isinstance(self.datasets[k], DataDescription):
                            pass
                        else:
                            msg = "Expected type for key:" + k + " is either None or of type DataDescription"
                            logger.error(msg)
                            raise ValueError(msg)
            else:
                msg = "dataset_obj must be of type dict with minium keys of train and clean_test"
                logger.error(msg)
                raise ValueError(msg)
        else:
            # check types
            if type(self.experiment_path) != str:
                raise TypeError("Expected type 'string' for argument 'experiment_path', "
                                "instead got type: {}".format(type(self.experiment_path)))
            for fn in self.train_file:
                if type(fn) != str:
                    raise TypeError("Expected string or Iterable[string] for argument 'train_file', "
                                    "instead got type: {}".format(type(fn)))
            if type(self.clean_test_file) != str:
                raise TypeError("Expected type 'string' for argument 'clean_test_file', "
                                "instead got type: {}".format(type(self.clean_test_file)))
            if self.triggered_test_file is not None and type(self.triggered_test_file) != str:
                raise TypeError("Expected type 'string' for argument 'triggered_test_file', "
                                "instead got type: {}".format(type(self.triggered_test_file)))
            if not callable(self.data_transform):
                raise TypeError("Expected a function for argument 'data_transform', "
                                "instead got type: {}".format(type(self.data_transform)))
            if not callable(self.label_transform):
                raise TypeError("Expected a function for argument 'label_transform', "
                                "instead got type: {}".format(type(self.label_transform)))
            if not callable(self.data_loader) and type(self.data_loader) != str:
                raise TypeError("Expected a function or string for argument 'data_loader', "
                                "instead got type: {}".format(type(self.data_loader)))
            if not type(self.shuffle_train) == bool:
                raise TypeError("Expected type 'bool' for argument 'shuffle_train', "
                                "instead got type: {}".format(type(self.shuffle_train)))
            if not type(self.shuffle_clean_test) == bool:
                raise TypeError("Expected type 'bool' for argument 'shuffle_clean_test', "
                                "instead got type: {}".format(type(self.shuffle_clean_test)))
            if not type(self.shuffle_triggered_test) == bool:
                raise TypeError("Expected type 'bool' for argument 'shuffle_triggered_test', "
                                "instead got type: {}".format(type(self.shuffle_triggered_test)))

            # check if files and directories exist
            if not os.path.isdir(self.experiment_path):
                raise FileNotFoundError("{} directory was not found...".format(self.experiment_path))
            for f in self.train_file:
                if not os.path.isfile(os.path.join(self.experiment_path, f)):
                    raise FileNotFoundError("{} training file was not found".format(f))
            if not os.path.isfile(os.path.join(self.experiment_path, self.clean_test_file)):
                raise FileNotFoundError("{} clean test file file was not found".format(self.clean_test_file))
            if self.triggered_test_file is not None and \
                    not os.path.isfile(os.path.join(self.experiment_path, self.triggered_test_file)):
                raise FileNotFoundError("{} triggered test file file was not found".format(self.triggered_test_file))

            # check if training data is empty
            for f in self.train_file:
                train_path = os.path.join(self.experiment_path, f)
                train_df = pd.read_csv(train_path)
                if len(train_df) == 0:
                    err_msg = "'train_file' {} is empty".format(train_path)
                    logger.error(err_msg)
                    raise RuntimeError(err_msg)

            clean_test_path = os.path.join(self.experiment_path, self.clean_test_file)
            clean_test_df = pd.read_csv(clean_test_path)
            if len(clean_test_df) == 0:
                err_msg = "'clean_test_file' is empty"
                logger.error(err_msg)
                raise RuntimeError(err_msg)

            if not isinstance(self.data_type, str):
                msg = "data_type argument must be one of the following: " + str(VALID_DATA_TYPES)
                logger.error(msg)
                raise ValueError(msg)
            else:
                if self.data_type not in VALID_DATA_TYPES:
                    msg = "Unsupported data_type argument provided"
                    logger.error(msg)
                    raise ValueError(msg)




"""
File describes data description classes, which contain specific information that may be used
in order to instantiate an architecture
"""


class DataDescription:
    """
    Generic Data Description class from which all specific data type data descriptors
    """
    pass


class CSVTextDatasetDesc(DataDescription):
    """
    Information potentially relevant to instantiating models to process text data
    """
    def __init__(self, vocab_size, unk_idx, pad_idx):
        self.vocab_size = vocab_size
        self.unk_idx = unk_idx
        self.pad_idx = pad_idx


class CSVImageDatasetDesc(DataDescription):
    """
    Information potentially relevant to instantiating models to process image data
    """
    def __init__(self, num_samples, shuffled, num_classes):
        self.num_samples = num_samples
        self.shuffled = shuffled
        self.num_classes = num_classes



""" Defines valid devices on which models can be trained """
VALID_DEVICES = ['cpu', 'cuda']

""" Defines valid loss functions which can be specified when configuring an optimizer implementing the 
OptimizerInterface"""
VALID_LOSS_FUNCTIONS = ['cross_entropy_loss', 'BCEWithLogitsLoss']

""" Defines valid optimization algorithms which can be specified when configuring an optimizer implementing the 
OptimizerInterface """
VALID_OPTIMIZERS = ['adam']

""" Defines the valid types of data that the modelgen pipeline can handle
"""
VALID_DATA_TYPES = ['image', 'text', 'custom']



from abc import ABC, abstractmethod
from typing import Sequence
import logging

import torch.nn

from .data_manager import CSVDataset
from .training_statistics import EpochStatistics

logger = logging.getLogger(__name__)


class OptimizerInterface(ABC):
    """Object that performs training and testing of TrojAI models."""
    @abstractmethod
    def train(self, model: torch.nn.Module, data: CSVDataset, train_val_split: float, torch_dataloader_kwargs) \
            -> (torch.nn.Module, Sequence[EpochStatistics]):
        """
        Train the given model using parameters in self.training_params
        :param model: (torch.nn.Module) The untrained Pytorch model
        :param data: (CSVDataset) Object containing training data, output 0 from TrojaiDataManager.load_data()
        :param train_val_split: (float) percentage of data that should be used for validation
        :param torch_dataloader_kwargs: additional arguments to pass to PyTorch's DataLoader class
        :return: (torch.nn.Module, EpochStatistics) trained model and sequence of EpochStatistics objects (one for
            each epoch).
        """
        pass

    @abstractmethod
    def test(self, model, clean_test_data, triggered_test_data, torch_dataloader_kwargs) -> dict:
        """
        Perform whatever tests desired on the model with clean data and triggered data, return a dictionary of results.
        :param model: (torch.nn.Module) Trained Pytorch model
        :param clean_test_data: (CSVDataset) Object containing clean test data
        :param triggered_test_data: (CSVDataset or None) Object containing triggered test data, None if triggered data
            was not provided for testing
        :param torch_dataloader_kwargs: additional arguments to pass to PyTorch's DataLoader class
        :return: (dict) Dictionary of test accuracy results.
            Required key, value pairs are:
                clean_accuracy: (float in [0, 1]) classification accuracy on clean data
                clean_n_total: (int) number of examples in clean test set
            The following keys are optional, but should be used if triggered test data was provided
                triggered_accuracy: (float in [0, 1]) classification accuracy on triggered data
                triggered_n_total: (int) number of examples in triggered test set

        NOTE: This list may be augmented in the future to allow for additional test data collection.
        """
        pass

    @abstractmethod
    def get_device_type(self) -> str:
        """
        Return a string representation of the type of device used by the optimizer to train the model.
        """
        pass

    @abstractmethod
    def get_cfg_as_dict(self) -> dict:
        """
        Return a dictionary with key/value pairs that describe the parameters used to train the model.
        """
        pass

    @abstractmethod
    def __deepcopy__(self, memodict={}):
        """
        Required for training on clusters. Return a deep copy of the optimizer.
        """
        pass

    @abstractmethod
    def __eq__(self, other):
        """
        Required for training on clusters. Define how to chech if two optimizers are equal.
        """
        pass

    @abstractmethod
    def __str__(self):
        pass

    @abstractmethod
    def save(self, fname: str) -> None:
        """
        Save the optimizer to a file
        :param fname - the filename to save the optimizer to
        """
        pass

    @staticmethod
    @abstractmethod
    def load(fname: str):
        """
        Load an optimizer from disk and return it
        :param fname: the filename where the optimizer is serialized
        :return: The loaded optimizer
        """
        pass
